

"Functions implementing widget editing"

import re, html, json

from skipole import skilift
from skipole.skilift import fromjson, editsection, editpage, editwidget, versions

from .. import utils
from skipole import FailPage, ValidateError, ServerError, GoTo

# a search for anything none-alphanumeric and not an underscore
_AN = re.compile('[^\w]')


def _field_name(widget, field_argument):
    "Returns a field name"
    if "set_names" not in widget:
        return field_argument
    name_dict = widget["set_names"]
    if field_argument in name_dict:
        return name_dict[field_argument]
    return field_argument


def _field_value(widget, field_argument):
    "Returns value,string value"
    value = widget["fields"][field_argument]
    if value is None:
        field_value = ''
    elif isinstance(value, list):
        if value:
            field_value = ','.join(str(val) for val in value)
        else:
            field_value = ''
    else:
        field_value = str(value)
    return value, field_value


def _field_ref(widgetdescription, field_argument):
    "Returns a field textblock reference string"
    if field_argument == 'show':
        return 'widgets.show'
    elif field_argument == 'widget_class':
        return 'widgets.widget_class'
    elif field_argument == 'widget_style':
        return 'widgets.widget_style'
    elif field_argument == 'show_error':
        return 'widgets.show_error'
    elif field_argument == 'clear_error':
        return 'widgets.clear_error'
    else:
        return ".".join(("widgets", widgetdescription.modulename, widgetdescription.classname, field_argument))


def retrieve_widget(skicall):
    "Fills in the edit a widget page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # get the widget name
    if ("left_nav","navbuttons","nav_links") in call_data:
        # should be submitted as widgetname from left navigation links
        widget_name = call_data["left_nav","navbuttons","nav_links"]
    elif 'widget_name' in call_data:
        widget_name = call_data['widget_name']
    elif 'part_tuple' in call_data:
        # called from dom table, via responder that finds what is being edited
        # and has set it into part_tuple
        part_tuple = call_data['part_tuple']
        widget_name = part_tuple.name
    else:
        raise FailPage(message="Invalid widget")

    if not widget_name:
        raise FailPage(message="Invalid widget")

    # and this is the widget to be edited, it is now set into session data
    call_data['widget_name'] = widget_name

    # Fill in header
    page_data[("adminhead","page_head","large_text")] = "Widget " + widget_name

    project = call_data['editedprojname']

    section_name = None
    pagenumber = None
    if 'section_name' in call_data:
        section_name = call_data['section_name']
    elif 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message="No section or page given")

    try:
        if section_name:
            widgetdescription = editwidget.section_widget_description(project, section_name, call_data['schange'], widget_name)
            widget =  editwidget.section_widget(project, section_name, call_data['schange'], widget_name)
        else:
            widgetdescription = editwidget.page_widget_description(project, pagenumber, call_data['pchange'], widget_name)
            widget = editwidget.page_widget(project, pagenumber, call_data['pchange'], widget_name)
    except ServerError as e:
        raise FailPage(e.message)
    

    page_data[('widget_type','para_text')] = "This widget is of type %s.%s." % (widgetdescription.modulename, widgetdescription.classname)
    page_data[('widget_textblock','textblock_ref')] = ".".join(("widgets", widgetdescription.modulename, widgetdescription.classname))
    page_data[('widget_name','input_text')] = widget_name
    page_data[('widget_brief','input_text')] = widgetdescription.brief

    # widgetdescription.fields_single is a list of namedtuples, each inner namedtuple representing a field
    # with items ['field_arg', 'field_type', 'valdt', 'jsonset', 'cssclass', 'cssstyle']

    args = widgetdescription.fields_single
    arg_list = widgetdescription.fields_list
    arg_table = widgetdescription.fields_table 
    arg_dict = widgetdescription.fields_dictionary

    if arg_list or arg_table or arg_dict:
        page_data[('args_multi','show')] = True
    else:
        page_data[('args_multi','show')] = False


    # args is shown on a LinkTextBlockTable2

    # contents row is
    # col 0 is the visible text to place in the link,
    # col 1 is the get field of the link
    # col 2 is the second get field of the link
    # col 3 is text appearing in the second table column
    # col 4 is the reference string of a textblock to appear the third table column
    # col 5 is text to appear if the reference cannot be found in the database
    # col 6 normally empty string, if set to text it will replace the textblock

    args_valdt = False

    args_content = []
    if args:
        for arg in args:
            name = _field_name(widget, arg.field_arg)
            ref = _field_ref(widgetdescription, arg.field_arg)
            if arg.valdt:   
                name = "* " + name
                args_valdt = True
            # field value
            value,field_value = _field_value(widget, arg.field_arg)
            if len(field_value) > 20:
                field_value = field_value[:18]
                field_value += '...'
            arg_row = [ name, arg.field_arg, '',field_value, ref, 'No description for %s' % (ref,), '']
            args_content.append(arg_row)
        page_data[('args','link_table')] = args_content
    else:
        page_data[('args','show')] = False
        page_data[('args_description','show')] = False


    # arg_list, arg_table and arg_dict are shown on LinkTextBlockTable widgets

    # contents row is
    # col 0 is the visible text to place in the link,
    # col 1 is the get field of the link
    # col 2 is the second get field of the link
    # col 3 is the reference string of a textblock to appear in the column adjacent to the link
    # col 4 is text to appear if the reference cannot be found in the database
    # col 5 normally empty string, if set to text it will replace the textblock


    arg_list_content = []
    if arg_list:
        for arg in arg_list:
            name = _field_name(widget, arg.field_arg)
            ref = _field_ref(widgetdescription, arg.field_arg)
            if arg.valdt:
                name = "* " + name
                args_valdt = True
            arg_row = [ name, arg.field_arg, '', ref, 'No description for %s' % (ref,), '']
            arg_list_content.append(arg_row)
        page_data[('arg_list','link_table')] = arg_list_content
    else:
        page_data[('arg_list','show')] = False
        page_data[('arg_list_description','show')] = False

    arg_table_content = []
    if arg_table:
        for arg in arg_table:
            name = _field_name(widget, arg.field_arg)
            ref = _field_ref(widgetdescription, arg.field_arg)
            if arg.valdt:
                name = "* " + name
                args_valdt = True
            arg_row = [ name, arg.field_arg, '', ref, 'No description for %s' % (ref,), '']
            arg_table_content.append(arg_row)
        page_data[('arg_table','link_table')] = arg_table_content
    else:
        page_data[('arg_table','show')] = False
        page_data[('arg_table_description','show')] = False

    arg_dict_content = []
    if arg_dict:
        for arg in arg_dict:
            name = _field_name(widget, arg.field_arg)
            ref = _field_ref(widgetdescription, arg.field_arg)
            if arg.valdt:
                name = "* " + name
                args_valdt = True
            arg_row = [ name, arg.field_arg, '', ref, 'No description for %s' % (ref,), '']
            arg_dict_content.append(arg_row)
        page_data[('arg_dict','link_table')] = arg_dict_content
    else:
        page_data[('arg_dict','show')] = False
        page_data[('arg_dict_description','show')] = False

    page_data[('args_valdt','show')] = args_valdt

    # display the widget html
    page_data[('widget_code','pre_text')] = widgetdescription.illustration

    if widgetdescription.containers:
        page_data[('containerdesc','show')] = True

    # remove any unwanted fields from session call_data
    if 'container' in call_data:
        del call_data['container']
    if 'location' in call_data:
        del call_data['location']
    if 'part' in call_data:
        del call_data['part']
    if 'field_arg' in call_data:
        del call_data['field_arg']
    if 'validx' in call_data:
        del call_data['validx']


def set_widget_params(skicall):
    "Sets widget name and brief"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    section_name = None
    pagenumber = None
    if 'section_name' in call_data:
        section_name = call_data['section_name']
    elif 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message="No section or page given")

    if 'widget_name' in call_data:
        widget_name = call_data['widget_name']
    else:
        raise FailPage(message="Widget not identified")

    new_name = None
    brief = None

    if 'new_widget_name' in call_data:
        new_name = call_data['new_widget_name']
    elif 'widget_brief' in call_data:
        brief = call_data['widget_brief']
    else:
        raise FailPage(message="No new name or brief given")
    
    try:
        if section_name:
            if new_name:
                call_data['schange'] = editwidget.rename_section_widget(project, section_name, call_data['schange'], widget_name, new_name)
                call_data['status'] = "Widget name changed"
                call_data['widget_name'] = new_name
            else:
                call_data['schange'] = editwidget.new_brief_in_section_widget(project, section_name, call_data['schange'], widget_name, brief)
                call_data['status'] = "Widget brief changed"
        else:
            if new_name:
                call_data['pchange'] = editwidget.rename_page_widget(project, pagenumber, call_data['pchange'], widget_name, new_name)
                call_data['status'] = "Widget name changed"
                call_data['widget_name'] = new_name
            else:
                call_data['pchange'] = editwidget.new_brief_in_page_widget(project, pagenumber, call_data['pchange'], widget_name, brief)
                call_data['status'] = "Widget brief changed"
    except ServerError as e:
        raise FailPage(e.message)


def retrieve_editfield(skicall):
    "Fills in the edit a widget field page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    section_name = None
    pagenumber = None
    if 'section_name' in call_data:
        section_name = call_data['section_name']
    elif 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message="No section or page given")

    if 'widget_name' in call_data:
        widget_name = call_data['widget_name']
    else:
        raise FailPage(message="Widget not identified")

    if 'field_arg' in call_data:
        field_arg = call_data['field_arg']
    else:
        raise FailPage("Field not identified")

    try:
        if section_name:
            widgetdescription = editwidget.section_widget_description(project, section_name, call_data['schange'], widget_name)
            widget =  editwidget.section_widget(project, section_name, call_data['schange'], widget_name)
        else:
            widgetdescription = editwidget.page_widget_description(project, pagenumber, call_data['pchange'], widget_name)
            widget = editwidget.page_widget(project, pagenumber, call_data['pchange'], widget_name)
    except ServerError as e:
        raise FailPage(e.message)

    page_data[('widget_type','para_text')] = "Widget type : %s.%s" % (widgetdescription.modulename, widgetdescription.classname)
    page_data[('widget_name','para_text')] = "Widget name : %s" % (widget_name,)
    page_data[('field_type','para_text')] = "Field type : %s" % (field_arg,)

    # widgetdescription.fields_single is a list of namedtuples, each inner namedtuple representing a field
    # with items ['field_arg', 'field_type', 'valdt', 'jsonset', 'cssclass', 'cssstyle']

    # create dictionaries of {field_arg : namedtuples }
    fields_single = { arg.field_arg:arg for arg in widgetdescription.fields_single }
    fields_list = { arg.field_arg:arg for arg in widgetdescription.fields_list }
    fields_table = { arg.field_arg:arg for arg in widgetdescription.fields_table }
    fields_dictionary = { arg.field_arg:arg for arg in widgetdescription.fields_dictionary }

    if field_arg in fields_single:
        field_datalist = fields_single[field_arg]
    elif field_arg in fields_list:
        field_datalist = fields_list[field_arg]
    elif field_arg in fields_table:
        field_datalist = fields_table[field_arg]
    elif field_arg in fields_dictionary:
        field_datalist = fields_dictionary[field_arg]
    else:
        raise FailPage("Field not identified")

    if field_datalist.jsonset:
        page_data[('json_enabled','para_text')] = "JSON Enabled : Yes"
    else:
        page_data[('json_enabled','para_text')] = "JSON Enabled : No"

    if field_arg in fields_single:
        if field_datalist.cssclass or field_datalist.cssstyle:
            default_value = skilift.fromjson.get_widget_default_field_value(project, widgetdescription.modulename, widgetdescription.classname, field_arg)
            if default_value:
                page_data[('field_default','para_text')] = "Default value : " + default_value
                page_data[('field_default','show')] = True

    field_name = _field_name(widget, field_arg)
    page_data[("adminhead","page_head","large_text")] = "(\'%s\',\'%s\')" % (widget_name, field_name)
    page_data[('show_field_name','para_text')] = "Field name : %s" % (field_name,)

    value, field_value = _field_value(widget, field_arg)

    # show the textblock description with .full, or if it doesnt exist, without the .full
    ref = _field_ref(widgetdescription, field_arg)

    full_textref = ref + '.full'   # the field reference string
    adminaccesstextblocks = skilift.get_accesstextblocks(skicall.project)

    if adminaccesstextblocks.textref_exists(full_textref):
        page_data[('widget_field_textblock','textblock_ref')] = full_textref
    else:
        page_data[('widget_field_textblock','textblock_ref')] = ref
    page_data[('field_name','input_text')] = field_name

    replace_strings = [widget_name+'\",\"'+field_name]

    if field_arg in fields_single:
        if field_datalist.field_type == 'boolean':
            page_data[("field_submit",'show')] = True
            page_data[("boolean_field_value", "radio_checked")] = value
        else:
            page_data[("field_value",'show')] = True
            page_data[("field_value",'input_text')] = field_value
        if field_datalist.cssclass or field_datalist.cssstyle:
            # add button to set given css class or style to defaults.json
            page_data[("css_default_desc",'show')] = True
            page_data[("set_field_default",'show')] = True
        else:
            page_data[("css_default_desc",'show')] = False
            page_data[("set_field_default",'show')] = False

        page_data[("show_field_value",'show')] = True
        page_data[("show_field_value",'para_text')] = "Field value : %s" % (field_value,)
        page_data[("widget_args_desc",'show')] = True
        page_data[("widget_args_desc",'replace_strings')] = replace_strings
    elif field_arg in fields_list:
        page_data[("widget_arg_list_desc",'show')] = True
        page_data[("widget_arg_list_desc",'replace_strings')] = replace_strings
        page_data[("css_default_desc",'show')] = False
        page_data[("set_field_default",'show')] = False
    elif field_arg in fields_table:
        page_data[("widget_arg_table_desc",'show')] = True
        page_data[("widget_arg_table_desc",'replace_strings')] = replace_strings
        page_data[("css_default_desc",'show')] = False
        page_data[("set_field_default",'show')] = False
    elif field_arg in fields_dictionary:
        page_data[("widget_arg_dict_desc",'show')] = True
        page_data[("widget_arg_dict_desc",'replace_strings')] = replace_strings
        page_data[("css_default_desc",'show')] = False
        page_data[("set_field_default",'show')] = False

    # Show validators
    if field_datalist.valdt:
        page_data[("validators_desc",'show')] = True
        page_data[("validators_desc2",'show')] = True
        page_data[("add_validator",'show')] = True
        # create the contents for the validator_table
        contents = []
        if ("validators" in widget) and (field_arg in widget["validators"]):
            val_list = widget["validators"][field_arg]
            page_data["validator_table:show"] = True
            max_validator_index = len(val_list) - 1
            for index,validator in enumerate(val_list):
                if index:
                    up = True
                else:
                    # first item (index zero) has no up button
                    up = False
                if index < max_validator_index:
                    down = True
                else:
                    # last item has no down button
                    down = False
                table_pos =  str(index)
                contents.append([validator['class'], table_pos, table_pos, table_pos, table_pos, True, up, down, True])
            page_data["validator_table:contents"] = contents

    # set field_arg into session call_data
    call_data['field_arg'] = field_arg

    if 'validx' in call_data:
        del call_data['validx']


def set_field_name(skicall):
    "Sets a widget field name"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    section_name = None
    pagenumber = None
    if 'section_name' in call_data:
        section_name = call_data['section_name']
    elif 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message="No section or page given")

    if 'widget_name' in call_data:
        widget_name = call_data['widget_name']
    else:
        raise FailPage(message="Widget not identified")

    if 'field_arg' in call_data:
        field_arg = call_data['field_arg']
    else:
        raise FailPage("Field not identified")

    if 'field_name' in call_data:
        field_name = call_data['field_name']
    else:
        raise FailPage("New field name not identified")

    try:
        if section_name:
            call_data['schange'] = editwidget.set_widget_field_name_in_section(project, section_name, call_data['schange'], widget_name, field_arg, field_name)
        else:
            call_data['pchange'] = editwidget.set_widget_field_name_in_page(project, pagenumber, call_data['pchange'], widget_name, field_arg, field_name)
    except ServerError as e:
        raise FailPage(e.message)

    call_data['status'] = "Field name changed"


def set_field_value(skicall):
    "Sets a widget field value"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    section_name = None
    pagenumber = None
    if 'section_name' in call_data:
        section_name = call_data['section_name']
    elif 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message="No section or page given")

    if 'widget_name' in call_data:
        widget_name = call_data['widget_name']
    else:
        raise FailPage(message="Widget not identified")

    if 'field_arg' in call_data:
        field_arg = call_data['field_arg']
    else:
        raise FailPage("Field not identified")

    if 'field_value' in call_data:
        field_value = call_data['field_value']
    else:
        raise FailPage("New field value not identified")

    try:
        if section_name:
            call_data['schange'] = editwidget.set_widget_field_value_in_section(project, section_name, call_data['schange'], widget_name, field_arg, field_value)
        else:
            call_data['pchange'] = editwidget.set_widget_field_value_in_page(project, pagenumber, call_data['pchange'], widget_name, field_arg, field_value)
    except ServerError as e:
        raise FailPage(e.message)

    call_data['status'] = "Field value changed"


def set_field_default(skicall):
    "Sets a widget field default value"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    section_name = None
    pagenumber = None
    if 'section_name' in call_data:
        section_name = call_data['section_name']
    elif 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message="No section or page given")

    if 'widget_name' in call_data:
        widget_name = call_data['widget_name']
    else:
        raise FailPage(message="Widget not identified")

    if 'field_arg' in call_data:
        field_arg = call_data['field_arg']
    else:
        raise FailPage("Field not identified")

    try:
        if section_name:
            widgetdescription = editwidget.section_widget_description(project, section_name, call_data['schange'], widget_name)
            widget =  editwidget.section_widget(project, section_name, call_data['schange'], widget_name)
        else:
            widgetdescription = editwidget.page_widget_description(project, pagenumber, call_data['pchange'], widget_name)
            widget = editwidget.page_widget(project, pagenumber, call_data['pchange'], widget_name)
    except ServerError as e:
        raise FailPage(e.message)

    fields_single = { arg.field_arg:arg for arg in widgetdescription.fields_single }
    if field_arg not in fields_single:
        raise FailPage("Cannot set a default value on this field")
    field_info = fields_single[field_arg]

    value, str_value = _field_value(widget, field_arg)

    if field_info.cssclass or field_info.cssstyle:
        # set the default value
        result = fromjson.save_widget_default_field_value(project,
                                                          widgetdescription.modulename,
                                                          widgetdescription.classname,
                                                          field_arg,
                                                          str_value)
        if result:
            if str_value:
                call_data['status'] = "Field default value set to %s" % (str_value,)
            else:
                call_data['status'] = "Field default value removed"
            return

    raise FailPage("Unable to set default")



def retrieve_container(skicall):
    "Edits a widget container"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # remove any unwanted fields from session call_data
    if 'location' in call_data:
        del call_data['location']
    if 'part' in call_data:
        del call_data['part']

    # get data
    if ("left_nav","navbuttons","nav_links") in call_data:
        # should be submitted as widgetname-containernumber
        widget_container = call_data["left_nav","navbuttons","nav_links"].split("-")
        if len(widget_container) != 2:
            raise FailPage(message="Invalid container")
        widget_name = widget_container[0]
        try:
            container = int(widget_container[1])
        except Exception:
            raise FailPage(message="Invalid container")
    elif ('widget_name' in call_data) and ('container' in call_data):
        widget_name = call_data['widget_name']
        container = call_data['container']
    else:
        raise FailPage(message="Invalid container")

    # and this is the container to be edited, it is now set into session data
    call_data['widget_name'] = widget_name
    call_data['container'] = container

    project = call_data['editedprojname']

    section_name = None
    pagenumber = None
    if 'section_name' in call_data:
        section_name = call_data['section_name']
    elif 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message="No section or page given")

    try:
        if section_name:
            widgetdescription = editwidget.section_widget_description(project, section_name, call_data['schange'], widget_name)
            containerinfo = editwidget.container_in_section(project, section_name, call_data['schange'], widget_name, container)
        else:
            widgetdescription = editwidget.page_widget_description(project, pagenumber, call_data['pchange'], widget_name)
            containerinfo = editwidget.container_in_page(project, pagenumber, call_data['pchange'], widget_name, container)
    except ServerError as e:
        raise FailPage(e.message)

    # containerinfo is a namedtuple ('container', 'container_ref', 'empty')

    # Fill in header
    page_data[("adminhead","page_head","large_text")] = "Widget " + widget_name + " container: " + str(container)

    # so header text and navigation done, now continue with the page contents
    page_data[('container_description','textblock_ref')] = ".".join(("widgets",widgetdescription.modulename, widgetdescription.classname, "container" + str(container)))
    if containerinfo.empty:
        # empty container
        page_data[('further_description','para_text')] = "The container is empty. Please choose an item to insert."
        # do not show the container table
        page_data[('editdom', 'show')] = False
        # table contents do not include Insert a section
        page_data[("insertlist","links")] = [
                ["Insert text", "inserttext", ""],
                ["Insert a TextBlock", "insert_textblockref", ""],
                ["Insert html symbol", "insertsymbol", ""],
                ["Insert comment", "insertcomment", ""],
                ["Insert an html element", "part_insert", ""],
                ["Insert a Widget", "list_widget_modules", ""]
                ]
        if not section_name:
            # going into a page, so a sectionplaceholder can be added
            page_data[("insertlist","links")].append(["Insert a Section", "placeholder_insert", ""])
        # set location, where item is to be inserted
        call_data['location'] = (widget_name, container, (0,))
        return

    # part has content, do not show insertlist or upload button
    page_data[('further_description','para_text')] = "Choose an item to edit."
    page_data[('insertlist', 'show')] = False
    page_data['upload_description', 'show'] = False
    page_data['uploadpart', 'show'] = False

    # fill in the table
    call_data['location_string'] = widget_name
    retrieve_container_dom(skicall)

    # and do show the download button
    page_data['download_description', 'show'] = True
    page_data['containerdownload', 'show'] = True




def retrieve_container_dom(skicall):
    "this call fills in the container dom table"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    pagenumber = None
    section_name = None

    if "page_number" in call_data:
        pagenumber = call_data["page_number"]
    elif "section_name" in call_data:
        section_name = call_data["section_name"]
    else:
        raise FailPage(message = "No page or section given")

    # location_string is the widget name

    if 'location_string' in call_data:
        location_string = call_data['location_string']
    elif 'widget_name' in call_data:
        location_string = call_data['widget_name']
    else:
        raise FailPage(message = "widget_name not in call_data")

    if 'container' not in call_data:
        raise FailPage(message = "container not in call_data")
    container = call_data["container"]

    try:
        domcontents, dragrows, droprows = _container_domcontents(project, pagenumber, section_name, location_string, container)
    except ServerError as e:
        raise FailPage(message=e.message)

    # widget editdom,domtable is populated with fields

    #    dragrows: A two element list for every row in the table, could be empty if no drag operation
    #              0 - True if draggable, False if not
    #              1 - If 0 is True, this is data sent with the call wnen a row is dropped
    #    droprows: A two element list for every row in the table, could be empty if no drop operation
    #              0 - True if droppable, False if not
    #              1 - text to send with the call when a row is dropped here
    #    dropident: ident or label of target, called when a drop occurs which returns a JSON page

    #    cols: A three element list for every column in the table, must be given with empty values if no links
    #              0 - target HTML page link ident of buttons in each column, if col1 not present or no javascript
    #              1 - target JSON page link ident of buttons in each column,
    #              2 - session storage key 'ski_part'

    #    contents: A list for every element in the table, should be row*col lists
    #               0 - text string, either text to display or button text
    #               1 - A 'style' string set on the td cell, if empty string, no style applied
    #               2 - Is button? If False only text will be shown, not a button, button class will not be applied
    #                       If True a link to link_ident/json_ident will be set with button_class applied to it
    #               3 - The get field value of the button link, empty string if no get field

    # create the table

    page_data['editdom', 'domtable', 'contents']  = domcontents
    page_data['editdom', 'domtable', 'dragrows']  = dragrows
    page_data['editdom', 'domtable', 'droprows']  = droprows

    # for each column: html link, JSON link, storage key
    page_data['editdom', 'domtable', 'cols']  =  [    ['','',''],                                                 # tag name, no link
                                                      ['','',''],                                                 # brief, no link
                                                      ['no_javascript','move_up_in_container_dom',''],            # up arrow
                                                      ['no_javascript','move_up_right_in_container_dom',''],      # up right
                                                      ['no_javascript','move_down_in_container_dom',''],          # down
                                                      ['no_javascript','move_down_right_in_container_dom',''],    # down right
                                                      ['edit_container_dom','',''],                               # edit, html only
                                                      ['add_to_container_dom','',''],                             # insert/append, html only
                                                      ['no_javascript',44580,''],                                 # copy
                                                      ['no_javascript',44590,'ski_part'],                         # paste
                                                      ['no_javascript','remove_container_dom','']                 # remove
                                                   ]

    page_data['editdom', 'domtable', 'dropident'] = 'move_in_container_dom'



def _container_domcontents(project, pagenumber, section_name, location_string, container):
    "Return the info for domtable contents"

    parttext, partdict = fromjson.container_outline(project, pagenumber, section_name, location_string, container)

    # create first row of the table

    top_row_widget = "Widget %s" % location_string
    top_row_container = "Container %s" % container

    domcontents = [
                   [top_row_widget, '', False, '' ],
                   [top_row_container, '', False, '' ],
                   ['', '', False, '' ],
                   ['', '', False, '' ],
                   ['', '', False, '' ],
                   ['', '', False, '' ],
                   ['', '', False, '' ],
                   ['', '', False, '' ],
                   ['', '', False, '' ],
                   ['', '', False, '' ],
                   ['', '', False, '' ]
                ]
    # add further items to domcontents
    part_string_list = []
    part_loc = location_string + '-' + str(container)
    rows = utils.domtree(partdict, part_loc, domcontents, part_string_list)

    # for every row in the table
    dragrows = [[ False, '']]
    droprows = [[ True, part_loc]]

    # for each row
    if rows>1:
        for row in range(0, rows-1):
            dragrows.append( [ True, part_string_list[row]] )
            droprows.append( [ True, part_string_list[row]] )
    
    return domcontents, dragrows, droprows


def copy_container(skicall):
    "Gets container part and return it in page_data['localStorage'] with key ski_part for browser session storage"
    call_data = skicall.call_data
    page_data = skicall.page_data
    project = call_data['editedprojname']
    pagenumber = None
    section_name = None

    if "page_number" in call_data:
        pagenumber = call_data["page_number"]
    elif "section_name" in call_data:
        section_name = call_data["section_name"]
    else:
        raise FailPage(message = "No page or section given")
    if ('editdom', 'domtable', 'contents') not in call_data:
        raise FailPage(message = "item missing")

    part = call_data['editdom', 'domtable', 'contents']

    # so part is widget_name, container with location string of integers

    # create location which is a tuple or list consisting of three items:
    # a string of widget name
    # a container integer
    # a tuple or list of location integers
    location_list = part.split('-')
    # first item should be a string, rest integers
    if len(location_list) < 3:
        raise FailPage("Item has not been recognised")

    try:
        widget_name = location_list[0]
        container = int(location_list[1])
        location_integers = [ int(i) for i in location_list[2:]]
    except Exception:
        raise FailPage("Item has not been recognised")

    # location is a tuple of widget_name, container, tuple of location integers
    location = (widget_name, container, location_integers)

    # get a json string dump of the item outline, however change any Sections to Parts
    itempart, itemdict = fromjson.item_outline(project, pagenumber, section_name, location)
    if itempart == 'Section':
        jsonstring = json.dumps(['Part',itemdict], indent=0, separators=(',', ':'))
    else:
        jsonstring = json.dumps([itempart,itemdict], indent=0, separators=(',', ':'))
    page_data['localStorage'] = {'ski_part':jsonstring}
    call_data['status'] = 'Item copied, and can now be pasted.'


def paste_container(skicall):
    "Gets submitted json string and inserts it"
    call_data = skicall.call_data
    page_data = skicall.page_data
    project = call_data['editedprojname']
    pagenumber = None
    section_name = None

    if "page_number" in call_data:
        pagenumber = call_data["page_number"]
    elif "section_name" in call_data:
        section_name = call_data["section_name"]
    else:
        raise FailPage(message = "No page or section given")
    if ('editdom', 'domtable', 'contents') not in call_data:
        raise FailPage(message = "item missing")
    part = call_data['editdom', 'domtable', 'contents']
    if ('editdom', 'domtable', 'cols') not in call_data:
        raise FailPage(message = "item to paste missing")
    json_string = call_data['editdom', 'domtable', 'cols']

    # so part is widget_name, container with location string of integers

    # create location which is a tuple or list consisting of three items:
    # a string of widget name
    # a container integer
    # a tuple or list of location integers
    location_list = part.split('-')
    # first item should be a string, rest integers
    if len(location_list) < 3:
        raise FailPage("Item has not been recognised")

    try:
        widget_name = location_list[0]
        container = int(location_list[1])
        location_integers = [ int(i) for i in location_list[2:]]
    except Exception:
        raise FailPage("Item has not been recognised")

    # location is a tuple of widget_name, container, tuple of location integers
    location = (widget_name, container, location_integers)

    if section_name:
        call_data['schange'] = editsection.create_item_in_section(project, section_name, call_data['schange'], location, json_string)
    else:
        call_data['pchange'] = editpage.create_item_in_page(project, pagenumber, call_data['pchange'], location, json_string)

    domcontents, dragrows, droprows = _container_domcontents(project, pagenumber, section_name, widget_name, container)
    page_data['editdom', 'domtable', 'dragrows']  = dragrows
    page_data['editdom', 'domtable', 'droprows']  = droprows
    page_data['editdom', 'domtable', 'contents']  = domcontents



def downloadcontainer(skicall):
    "Gets container, and returns a json dictionary, this will be sent as an octet file to be downloaded"

    call_data = skicall.call_data
    page_data = skicall.page_data
    project = call_data['editedprojname']

    pagenumber = None
    section_name = None

    if "page_number" in call_data:
        pagenumber = call_data["page_number"]
    elif "section_name" in call_data:
        section_name = call_data["section_name"]
    else:
        raise FailPage(message = "No page or section given")

    if 'widget_name' in call_data:
        widget_name = call_data['widget_name']
    else:
        raise FailPage(message = "widget_name not in call_data")

    if 'container' not in call_data:
        raise FailPage(message = "container not in call_data")
    container = call_data["container"]

    parttext, part_dict = fromjson.container_outline(project, pagenumber, section_name, widget_name, container)
    # set contents into a div
    part_dict["hide_if_empty"] = False
    part_dict.move_to_end("hide_if_empty", last=False)
    part_dict["show"] = True
    part_dict.move_to_end("show", last=False)
    part_dict["brief"] = "From widget %s container %s" % (widget_name, container)
    part_dict.move_to_end("brief", last=False)
    part_dict["tag_name"] = "div"
    part_dict.move_to_end("tag_name", last=False)
    # set version and skipole as the first two items in the dictionary
    versions_tuple = versions(project)
    part_dict["skipole"] = versions_tuple.skipole
    part_dict.move_to_end('skipole', last=False)
    part_dict["version"] = versions_tuple.project
    part_dict.move_to_end('version', last=False)
    jsonstring = json.dumps(part_dict, indent=4, separators=(',', ':'))
    line_list = []
    n = 0
    for line in jsonstring.splitlines(True):
        binline = line.encode('utf-8')
        n += len(binline)
        line_list.append(binline)
    page_data['headers'] = [('content-type', 'application/octet-stream'), ('content-length', str(n))]
    return line_list


def back_to_parent_container(skicall):
    "Sets call_data['widget_name'] to parent_widget and call_data['container'] to parent_container"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    section_name = None
    pagenumber = None
    if 'section_name' in call_data:
        section_name = call_data['section_name']
    elif 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message="No section or page given")

    try:
        if section_name:
            widgetdescription = editwidget.section_widget_description(project, section_name, call_data['schange'], widget_name)
        else:
            widgetdescription = editwidget.page_widget_description(project, pagenumber, call_data['pchange'], widget_name)
    except ServerError as e:
        raise FailPage(e.message)

    utils.clear_call_data(call_data)

    if section_name:
        call_data['section_name'] = section_name
    else:
        call_data['page_number'] = pagenumber

    call_data['widget_name'] = widgetdescription.parent_widget
    call_data['container'] = widgetdescription.parent_container



def edit_container_dom(skicall):
    "Called by domtable to edit an item in a container"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    pagenumber = None
    section_name = None

    if "page_number" in call_data:
        pagenumber = call_data["page_number"]
    elif "section_name" in call_data:
        section_name = call_data["section_name"]
    else:
        raise FailPage(message = "No page or section given")
    if ('editdom', 'domtable', 'contents') not in call_data:
        raise FailPage(message = "item to edit missing")

    part = call_data['editdom', 'domtable', 'contents']

    # so part is widget_name, container with location string of integers

    # create location which is a tuple or list consisting of three items:
    # a string of widget name
    # a container integer
    # a tuple or list of location integers
    location_list = part.split('-')
    # first item should be a string, rest integers
    if len(location_list) < 3:
        raise FailPage("Item to edit has not been recognised")

    try:
        widget_name = location_list[0]
        container = int(location_list[1])
        location_integers = [ int(i) for i in location_list[2:]]
    except Exception:
        raise FailPage("Item to edit has not been recognised")


    part_tuple = skilift.part_info(project, pagenumber, section_name, [widget_name, container, location_integers])
    if part_tuple is None:
        raise FailPage("Item to edit has not been recognised")

    if part_tuple.name:
        # item to edit is a widget
        call_data['part_tuple'] = part_tuple
        raise GoTo(target = 54006, clear_submitted=True)
    if part_tuple.part_type == "Part":
        # edit the html part
        call_data['part_tuple'] = part_tuple
        raise GoTo(target = 53007, clear_submitted=True)
    if part_tuple.part_type == "ClosedPart":
        # edit the html closed part
        call_data['part_tuple'] = part_tuple
        raise GoTo(target = 53007, clear_submitted=True)
    if part_tuple.part_type == "HTMLSymbol":
        # edit the symbol
        call_data['part_tuple'] = part_tuple
        raise GoTo(target = 51107, clear_submitted=True)
    if part_tuple.part_type == "str":
        # edit the text
        call_data['part_tuple'] = part_tuple
        raise GoTo(target = 51017, clear_submitted=True)
    if part_tuple.part_type == "TextBlock":
        # edit the TextBlock
        call_data['part_tuple'] = part_tuple
        raise GoTo(target = 52017, clear_submitted=True)
    if part_tuple.part_type == "Comment":
        # edit the Comment
        call_data['part_tuple'] = part_tuple
        raise GoTo(target = 51207, clear_submitted=True)
    if (not section_name) and (part_tuple.part_type == "SectionPlaceHolder"):
        # edit the SectionPlaceHolder
        call_data['part_tuple'] = part_tuple
        raise GoTo(target = 55007, clear_submitted=True)

    raise FailPage("Item to edit has not been recognised")


def add_to_container_dom(skicall):
    """Called by domtable to either insert or append an item in a container
       sets page_data to populate the insert or append page and then go to appropriate template page"""

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    pagenumber = None
    section_name = None

    if "page_number" in call_data:
        pagenumber = call_data["page_number"]
    elif "section_name" in call_data:
        section_name = call_data["section_name"]
    else:
        raise FailPage(message = "No page or section given")
    if ('editdom', 'domtable', 'contents') not in call_data:
        raise FailPage(message = "item to append to missing")

    part = call_data['editdom', 'domtable', 'contents']

    # so part is widget_name, container with location string of integers

    # create location which is a tuple or list consisting of three items:
    # a string of widget name
    # a container integer
    # a tuple or list of location integers
    location_list = part.split('-')
    # first item should be a string, rest integers
    if len(location_list) < 3:
        raise FailPage("Item to append to has not been recognised")

    try:
        widget_name = location_list[0]
        container = int(location_list[1])
        location_integers = [ int(i) for i in location_list[2:]]
    except Exception:
        raise FailPage("Item to append to has not been recognised")

    # location is a tuple of widget_name, container, tuple of location integers
    location = (widget_name, container, location_integers)

    part_tuple = skilift.part_info(project, pagenumber, section_name, location)
    if part_tuple is None:
        raise FailPage("Item to append to has not been recognised")

    # goto either the install or append page, to add an item at this location
    call_data['location'] = location

    # Fill in menu of items, Part items have insert, others have append


    if (part_tuple.part_type == "Part") or (part_tuple.part_type == "Section"):
        # insert
        page_data[("adminhead","page_head","large_text")] = "Choose an item to insert"
        page_data[("insertlist","links")] = [
                                                ["Insert text", "inserttext", ""],
                                                ["Insert a TextBlock", "insert_textblockref", ""],
                                                ["Insert html symbol", "insertsymbol", ""],
                                                ["Insert comment", "insertcomment", ""],
                                                ["Insert an html element", "part_insert", ""],
                                                ["Insert a Widget", "list_widget_modules", ""]
                                            ]
        if not section_name:
            # going into a page, so a sectionplaceholder can be added
            page_data[("insertlist","links")].append(["Insert a Section", "placeholder_insert", ""])
        raise GoTo(target = '23609', clear_submitted=True)
    else:
        # append
        page_data[("adminhead","page_head","large_text")] = "Choose an item to append"
        page_data[("appendlist","links")] = [
                                                ["Append text", "inserttext", ""],
                                                ["Append a TextBlock", "insert_textblockref", ""],
                                                ["Append html symbol", "insertsymbol", ""],
                                                ["Append comment", "insertcomment", ""],
                                                ["Append an html element", "part_insert", ""],
                                                ["Append a Widget", "list_widget_modules", ""]
                                            ]
        if not section_name:
            # going into a page, so a sectionplaceholder can be added
            page_data[("appendlist","links")].append(["Append a Section", "placeholder_insert", ""])
        raise GoTo(target = '23509', clear_submitted=True)


def remove_container_dom(skicall):
    "Called by domtable to remove an item in a container"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    pagenumber = None
    section_name = None

    if "page_number" in call_data:
        pagenumber = call_data["page_number"]
    elif "section_name" in call_data:
        section_name = call_data["section_name"]
    else:
        raise FailPage(message = "No page or section given")
    if ('editdom', 'domtable', 'contents') not in call_data:
        raise FailPage(message = "item to remove missing")

    part = call_data['editdom', 'domtable', 'contents']

    # so part is widget_name, container with location string of integers

    # create location which is a tuple or list consisting of three items:
    # a string of widget name
    # a container integer
    # a tuple or list of location integers
    location_list = part.split('-')
    # first item should be a string, rest integers
    if len(location_list) < 3:
        raise FailPage("Item to remove has not been recognised")

    try:
        widget_name = location_list[0]
        container = int(location_list[1])
        location_integers = [ int(i) for i in location_list[2:]]
    except Exception:
        raise FailPage("Item to remove has not been recognised")

    # location is a tuple of widget_name, container, tuple of location integers
    location = (widget_name, container, location_integers)

    part_tuple = skilift.part_info(project, pagenumber, section_name, location)
    if part_tuple is None:
        raise FailPage("Item to remove has not been recognised")

    # prior to deleting, take a copy
    # get a json string dump of the item outline, however change any Sections to Parts
    itempart, itemdict = fromjson.item_outline(project, pagenumber, section_name, location)
    if itempart == 'Section':
        jsonstring = json.dumps(['Part',itemdict], indent=0, separators=(',', ':'))
    else:
        jsonstring = json.dumps([itempart,itemdict], indent=0, separators=(',', ':'))
    page_data['localStorage'] = {'ski_part':jsonstring}

    # remove the item using functions from skilift.editsection and skilift.editpage
    if pagenumber is None:
        # remove the item from a section
        try:
            call_data['schange'] = editsection.del_location(project, section_name, call_data['schange'], location)
        except ServerError as e:
            raise FailPage(message = e.message)
    else:
        # remove the item from a page
        try:
            call_data['pchange'] = editpage.del_location(project, pagenumber, call_data['pchange'], location)
        except ServerError as e:
            raise FailPage(message = e.message)

    # and get info to re-draw the table
    domcontents, dragrows, droprows = _container_domcontents(project, pagenumber, section_name, widget_name, container)

    # once item is deleted, no info on the item should be
    # left in call_data - this may not be required in future
    if 'location' in call_data:
        del call_data['location']
    if 'part' in call_data:
        del call_data['part']
    if 'part_loc' in call_data:
        del call_data['part_loc']

    call_data['container'] = container
    call_data['widget_name'] = widget_name
    call_data['status'] = 'Item copied and then deleted. Use paste to recover or move it.'

    # If deleting item has left container empty, return a full retrieve of the container page
    if len(domcontents) == 11:
        # 11 items indicates the top title row only, with no further contents
        raise GoTo("back_to_container")

    # otherwise just redraw the table
    page_data['editdom', 'domtable', 'dragrows']  = dragrows
    page_data['editdom', 'domtable', 'droprows']  = droprows
    page_data['editdom', 'domtable', 'contents']  = domcontents



def _item_to_move(call_data):
    "Gets the item to be moved"
    project = call_data['editedprojname']
    pagenumber = None
    section_name = None

    if "page_number" in call_data:
        pagenumber = call_data["page_number"]
    elif "section_name" in call_data:
        section_name = call_data["section_name"]
    else:
        raise FailPage(message = "No page or section given")
    if ('editdom', 'domtable', 'contents') not in call_data:
        raise FailPage(message = "item to move missing")

    part = call_data['editdom', 'domtable', 'contents']

    # so part is widget_name, container with location string of integers

    # create location which is a tuple or list consisting of three items:
    # a string of widget name
    # a container integer
    # a tuple or list of location integers
    location_list = part.split('-')
    # first item should be a string, rest integers
    if len(location_list) < 3:
        raise FailPage("Item to move has not been recognised")

    try:
        widget_name = location_list[0]
        container = int(location_list[1])
        location_integers = [ int(i) for i in location_list[2:]]
    except Exception:
        raise FailPage("Item to move has not been recognised")

    # location is a tuple of widget_name, container, tuple of location integers
    location = (widget_name, container, location_integers)
    call_data['container'] = container
    call_data['widget_name'] = widget_name
    try:
        part_tuple = skilift.part_info(project, pagenumber, section_name, location)
    except ServerError as e:
        raise FailPage(message = e.message)

    if part_tuple is None:
        raise FailPage("Item to move has not been recognised")
    return part_tuple


def move_up_in_container_dom(skicall):
    "Called by domtable to move an item in a container up"

    call_data = skicall.call_data
    page_data = skicall.page_data

    try:
        part_tuple = _item_to_move(call_data)
        location = part_tuple.location
        widget_name = location[0]
        container = int(location[1])
        location_integers = location[2]

        if (len(location_integers) == 1) and (location_integers[0] == 0):
            # at top, cannot be moved
            raise FailPage("Cannot be moved up")

        if location_integers[-1] == 0:
            # move up to next level
            new_location_integers = location_integers[:-1]
        else:
            # swap parts on same level
            new_location_integers = list(location_integers[:-1])
            new_location_integers.append(location_integers[-1] - 1)

        # after a move, location is wrong, so remove from call_data
        if 'location' in call_data:
            del call_data['location']
        if 'part' in call_data:
            del call_data['part']
        if 'part_top' in call_data:
            del call_data['part_top']
        if 'part_loc' in call_data:
            del call_data['part_loc']

        # move the item

        if part_tuple.section_name:
            # move the part in a section, using skilift.editsection.move_location(project, section_name, schange, from_location, to_location)
            call_data['schange'] = editsection.move_location(part_tuple.project, part_tuple.section_name, call_data['schange'], location, (widget_name, container, new_location_integers))
            domcontents, dragrows, droprows = _container_domcontents(part_tuple.project, None, part_tuple.section_name, widget_name, container)
        else:
            # move the part in a page, using skilift.editpage.move_location(project, pagenumber, pchange, from_location, to_location)
            call_data['pchange'] = editpage.move_location(part_tuple.project, part_tuple.pagenumber, call_data['pchange'], location, (widget_name, container, new_location_integers))
            domcontents, dragrows, droprows = _container_domcontents(part_tuple.project, part_tuple.pagenumber, None, widget_name, container)
    except ServerError as e:
        raise FailPage(message = e.message)

    # redraw the table
    page_data['editdom', 'domtable', 'dragrows']  = dragrows
    page_data['editdom', 'domtable', 'droprows']  = droprows
    page_data['editdom', 'domtable', 'contents']  = domcontents


def move_up_right_in_container_dom(skicall):
    "Called by domtable to move an item in a container up and to the right"

    call_data = skicall.call_data
    page_data = skicall.page_data

    try:
        part_tuple = _item_to_move(call_data)
        location = part_tuple.location
        widget_name = location[0]
        container = int(location[1])
        location_integers = location[2]

        if location_integers[-1] == 0:
            # at top of a part, cannot be moved
            raise FailPage("Cannot be moved up")
        new_parent_integers = list(location_integers[:-1])
        new_parent_integers.append(location_integers[-1] - 1)
        new_parent_location = (location[0], location[1], new_parent_integers)

        new_parent_tuple = skilift.part_info(part_tuple.project, part_tuple.pagenumber, part_tuple.section_name, new_parent_location)

        if new_parent_tuple is None:
            raise FailPage("Cannot be moved up")
        if new_parent_tuple.part_type != "Part":
            raise FailPage("Cannot be moved up")

        items_in_new_parent = len(skilift.part_contents(part_tuple.project, part_tuple.pagenumber, part_tuple.section_name, new_parent_location))

        new_location_integers =  tuple(new_parent_integers + [items_in_new_parent])

        # after a move, location is wrong, so remove from call_data
        if 'location' in call_data:
            del call_data['location']
        if 'part' in call_data:
            del call_data['part']
        if 'part_top' in call_data:
            del call_data['part_top']
        if 'part_loc' in call_data:
            del call_data['part_loc']

        # move the item

        if part_tuple.section_name:
            # move the part in a section, using skilift.editsection.move_location(project, section_name, schange, from_location, to_location)
            call_data['schange'] = editsection.move_location(part_tuple.project, part_tuple.section_name, call_data['schange'], location, (widget_name, container, new_location_integers))
            domcontents, dragrows, droprows = _container_domcontents(part_tuple.project, None, part_tuple.section_name, widget_name, container)
        else:
            # move the part in a page, using skilift.editpage.move_location(project, pagenumber, pchange, from_location, to_location)
            call_data['pchange'] = editpage.move_location(part_tuple.project, part_tuple.pagenumber, call_data['pchange'], location, (widget_name, container, new_location_integers))
            domcontents, dragrows, droprows = _container_domcontents(part_tuple.project, part_tuple.pagenumber, None, widget_name, container)
    except ServerError as e:
        raise FailPage(message = e.message)

    # redraw the table
    page_data['editdom', 'domtable', 'dragrows']  = dragrows
    page_data['editdom', 'domtable', 'droprows']  = droprows
    page_data['editdom', 'domtable', 'contents']  = domcontents


def move_down_in_container_dom(skicall):
    "Called by domtable to move an item in a container down"

    call_data = skicall.call_data
    page_data = skicall.page_data

    try:
        part_tuple = _item_to_move(call_data)
        location = part_tuple.location
        widget_name = location[0]
        container = int(location[1])
        location_integers = location[2]

        if len(location_integers) == 1:
            # Just at immediate level below top
            parent_location = (widget_name, container, ())
            items_in_parent = len(skilift.part_contents(part_tuple.project, part_tuple.pagenumber, part_tuple.section_name, parent_location))
            if location_integers[0] == (items_in_parent-1):
                # At end, cannot be moved
                raise FailPage("Cannot be moved down")
            new_location_integers = (location_integers[0]+2,)
        else:
            parent_integers = tuple(location_integers[:-1])
            parent_location = (widget_name, container, parent_integers)
            items_in_parent = len(skilift.part_contents(part_tuple.project, part_tuple.pagenumber, part_tuple.section_name, parent_location))
            if location_integers[-1] == (items_in_parent-1):
                # At end of a part, so move up a level
                new_location_integers = list(parent_integers[:-1])
                new_location_integers.append(parent_integers[-1] + 1)
            else:
                # just insert into current level
                new_location_integers = list(parent_integers)
                new_location_integers.append(location_integers[-1] + 2)

        # after a move, location is wrong, so remove from call_data
        if 'location' in call_data:
            del call_data['location']
        if 'part' in call_data:
            del call_data['part']
        if 'part_top' in call_data:
            del call_data['part_top']
        if 'part_loc' in call_data:
            del call_data['part_loc']

        # move the item

        if part_tuple.section_name:
            # move the part in a section, using skilift.editsection.move_location(project, section_name, schange, from_location, to_location)
            call_data['schange'] = editsection.move_location(part_tuple.project, part_tuple.section_name, call_data['schange'], location, (widget_name, container, new_location_integers))
            domcontents, dragrows, droprows = _container_domcontents(part_tuple.project, None, part_tuple.section_name, widget_name, container)
        else:
            # move the part in a page, using skilift.editpage.move_location(project, pagenumber, pchange, from_location, to_location)
            call_data['pchange'] = editpage.move_location(part_tuple.project, part_tuple.pagenumber, call_data['pchange'], location, (widget_name, container, new_location_integers))
            domcontents, dragrows, droprows = _container_domcontents(part_tuple.project, part_tuple.pagenumber, None, widget_name, container)
    except ServerError as e:
        raise FailPage(message = e.message)

    # redraw the table
    page_data['editdom', 'domtable', 'dragrows']  = dragrows
    page_data['editdom', 'domtable', 'droprows']  = droprows
    page_data['editdom', 'domtable', 'contents']  = domcontents


def move_down_right_in_container_dom(skicall):
    "Called by domtable to move an item in a container down and to the right"

    call_data = skicall.call_data
    page_data = skicall.page_data

    try:
        part_tuple = _item_to_move(call_data)
        location = part_tuple.location
        widget_name = location[0]
        container = int(location[1])
        location_integers = location[2]

        if len(location_integers) == 1:
            parent_location = (widget_name, container, ())
        else:
            parent_integers = list(location_integers[:-1])
            parent_location = (widget_name, container, parent_integers)
        items_in_parent = len(skilift.part_contents(part_tuple.project, part_tuple.pagenumber, part_tuple.section_name, parent_location))
        if location_integers[-1] == (items_in_parent-1):
            # At end of a block, cannot be moved
            raise FailPage("Cannot be moved down")
        new_parent_integers = list(location_integers[:-1])
        new_parent_integers.append(location_integers[-1] + 1)
        new_parent_location = (location[0], location[1], new_parent_integers)
        new_parent_tuple = skilift.part_info(part_tuple.project, part_tuple.pagenumber, part_tuple.section_name, new_parent_location)

        if new_parent_tuple is None:
            raise FailPage("Cannot be moved down")
        if not (new_parent_tuple.part_type == 'Part' or new_parent_tuple.part_type == 'Section'):
            raise FailPage("Cannot be moved down")

        new_location_integers = tuple(new_parent_integers+[0])

        # after a move, location is wrong, so remove from call_data
        if 'location' in call_data:
            del call_data['location']
        if 'part' in call_data:
            del call_data['part']
        if 'part_top' in call_data:
            del call_data['part_top']
        if 'part_loc' in call_data:
            del call_data['part_loc']

        # move the item

        if part_tuple.section_name:
            # move the part in a section, using skilift.editsection.move_location(project, section_name, schange, from_location, to_location)
            call_data['schange'] = editsection.move_location(part_tuple.project, part_tuple.section_name, call_data['schange'], location, (widget_name, container, new_location_integers))
            domcontents, dragrows, droprows = _container_domcontents(part_tuple.project, None, part_tuple.section_name, widget_name, container)
        else:
            # move the part in a page, using skilift.editpage.move_location(project, pagenumber, pchange, from_location, to_location)
            call_data['pchange'] = editpage.move_location(part_tuple.project, part_tuple.pagenumber, call_data['pchange'], location, (widget_name, container, new_location_integers))
            domcontents, dragrows, droprows = _container_domcontents(part_tuple.project, part_tuple.pagenumber, None, widget_name, container)
    except ServerError as e:
        raise FailPage(message = e.message)

    # redraw the table
    page_data['editdom', 'domtable', 'dragrows']  = dragrows
    page_data['editdom', 'domtable', 'droprows']  = droprows
    page_data['editdom', 'domtable', 'contents']  = domcontents


def move_in_container_dom(skicall):
    "Called by domtable to move an item in a container after a drag and drop"

    call_data = skicall.call_data
    page_data = skicall.page_data

    if ('editdom', 'domtable', 'dragrows') not in call_data:
        raise FailPage(message = "item to drop missing")
    editedprojname = call_data['editedprojname']
    pagenumber = None
    section_name = None

    if "page_number" in call_data:
        pagenumber = call_data["page_number"]
    elif "section_name" in call_data:
        section_name = call_data["section_name"]
    else:
        raise FailPage(message = "No page or section given")

    part_to_move = call_data['editdom', 'domtable', 'dragrows']

    # so part_to_move is widget name with container and location string of integers
    # create location which is a tuple or list consisting of three items:
    # a string of widget name
    # a container integer
    # a tuple or list of location integers
    location_list = part_to_move.split('-')
    # first item should be a string, rest integers
    if len(location_list) < 3:
        raise FailPage("Item to move has not been recognised")

    try:
        widget_name = location_list[0]
        container = int(location_list[1])
        location_to_move_integers = [ int(i) for i in location_list[2:]]
    except Exception:
        raise FailPage("Item to move has not been recognised")

    # location is a tuple of widget_name, container, tuple of location integers
    location_to_move = (widget_name, container, location_to_move_integers)
    call_data['container'] = container
    call_data['widget_name'] = widget_name


    # new target location

    target_part = call_data['editdom', 'domtable', 'droprows']

    # so target_part is widget name with location string of integers

    # create location which is a tuple or list consisting of three items:
    # a string of widget name
    # a container integer
    # a tuple or list of location integers

    location_list = target_part.split('-')
    # first item should be a string, rest integers
    if len(location_list) < 2:
        raise FailPage("target of move has not been recognised")
 
    if widget_name != location_list[0]:
        raise FailPage("Invalid move, widget name differs")

    if container != int(location_list[1]):
        raise FailPage("Invalid move, container number differs")

    if len(location_list) == 2:
            # At the container top row
            new_location_integers = [0]
    else:
        try:
            target_location_integers = [ int(i) for i in location_list[2:]]
        except Exception:
            raise FailPage("Invalid move, location not accepted")

        # location is a tuple of widget_name, container, tuple of location integers
        target_location = (widget_name, container, target_location_integers)

        # get target part_tuple from project, pagenumber, section_name, target_location
        target_part_tuple = skilift.part_info(editedprojname, pagenumber, section_name, target_location)
        if target_part_tuple is None:
            raise FailPage("Target has not been recognised")

        if (target_part_tuple.part_type == "Part") or (target_part_tuple.part_type == "Section"):
            # insert
            if target_location_integers:
                new_location_integers = list(target_location_integers)
                new_location_integers.append(0)
            else:
                new_location_integers = [0]
        else:
            # append
            new_location_integers = list(target_location_integers)
            new_location_integers[-1] = new_location_integers[-1] + 1

    # after a move, location is wrong, so remove from call_data
    if 'location' in call_data:
        del call_data['location']
    if 'part' in call_data:
        del call_data['part']
    if 'part_top' in call_data:
        del call_data['part_top']
    if 'part_loc' in call_data:
        del call_data['part_loc']

    # move the item
    try:
        if section_name:
            # move the part in a section, using skilift.editsection.move_location(project, section_name, schange, from_location, to_location)
            call_data['schange'] = editsection.move_location(editedprojname, section_name, call_data['schange'], location_to_move, (widget_name, container, new_location_integers))
            domcontents, dragrows, droprows = _container_domcontents(editedprojname, None, section_name, widget_name, container)
        else:
            # move the part in a page, using skilift.editpage.move_location(project, pagenumber, pchange, from_location, to_location)
            call_data['pchange']= editpage.move_location(editedprojname, pagenumber, call_data['pchange'], location_to_move, (widget_name, container, new_location_integers))
            domcontents, dragrows, droprows = _container_domcontents(editedprojname, pagenumber, None, widget_name, container)
    except ServerError as e:
        raise FailPage(message = e.message)

    # redraw the table
    page_data['editdom', 'domtable', 'dragrows']  = dragrows
    page_data['editdom', 'domtable', 'droprows']  = droprows
    page_data['editdom', 'domtable', 'contents']  = domcontents






import sys, traceback

from functools import wraps
from importlib import import_module

from .ski import skiboot

from .ski.project_class_definition import WSGIApplication
from .ski.excepts import ValidateError, ServerError, GoTo, FailPage

version = skiboot.version()


def set_debug(mode):
    "If mode is True, this sets increased debug error messages to be displayed"
    skiboot.set_debug(mode)



# This 'use_submit_list' is available to wrap the submit_data function if required


def use_submit_list(submit_data):
    "Used to decorate submit_data to enable submit_list to define package,module,function"
    @wraps(submit_data)
    def submit_function(skicall):
        "This function replaces submit_data, if skicall.submit_list has two or more elements"
        if not skicall.submit_list:
            # do nothing, simply return the original submit_data
            return submit_data(skicall)
        if len(skicall.submit_list) < 2:
            # do nothing, simply return the original submit_data
            return submit_data(skicall)
        submitpath = ".".join(skicall.submit_list[:-1])
        try:
            submitmodule = import_module(submitpath)
        except Exception:
            if skiboot.get_debug():
                exc_type, exc_value, exc_traceback = sys.exc_info()
                str_list = traceback.format_exception(exc_type, exc_value, exc_traceback)
                message = ''
                for item in str_list:
                    message += item
                raise ServerError(message)
            raise ServerError("Unable to import project module defined as %s in submit list" % (submitpath,))
        # now obtain and run the specified function
        try:
            submitfunc = getattr(submitmodule, skicall.submit_list[-1])
        except Exception:
            raise ServerError("submit_list package %s found, but the required function %s is not recognised" % (submitpath, skicall.submit_list[-1]))
        return submitfunc(skicall)
    return submit_function








import html

from skipole import FailPage, ServerError
from skipole import skilift


def clear_call_data(call_data, keep=None):
    "Clears call data apart from set of required values, and anything in the keep list"
    required = ['editedprojname',
                'editedprojurl',
                'editedprojversion',
                'editedprojbrief',
                'adminproj',
                'extend_nav_buttons',
                'caller_ident']
    if keep:
        if isinstance(keep, str) and (keep not in required):
            required.append(keep)
        else:
            # presume keep is a list of items
            for item in keep:
                if item not in required:
                    required.append(item)
    temp_storage = {key:value for key,value in call_data.items() if key in required}
    call_data.clear()
    call_data.update(temp_storage)


def domtree(partdict, part_loc, contents, part_string_list, rows=1, indent=1):
    "Creates the contents of the domtable"

    # note: if in a container
    # part_loc = widget_name + '-' + container_number
    # otherwise part_loc = body, head, svg, section_name


    indent += 1
    padding = "padding-left : %sem;" % (indent,)
    u_r_flag = False
    last_row_at_this_level = 0

    parts = partdict['parts']

    # parts is a list of items
    last_index = len(parts)-1

    #Text   #characters..      #up  #up_right  #down  #down_right   #edit   #insert  #copy  #paste  #remove

    for index, part in enumerate(parts):
        part_location_string = part_loc + '-' + str(index)
        part_string_list.append(part_location_string)
        rows += 1
        part_type, part_dict = part
        # the row text
        if part_type == 'Widget' or part_type == 'ClosedWidget':
            part_name = 'Widget ' + part_dict['name']
            if len(part_name)>40:
                part_name = part_name[:35] + '...'
            contents.append([part_name, padding, False, ''])
            part_brief = html.escape(part_dict.get('brief',''))
            if len(part_brief)>40:
                part_brief = part_brief[:35] + '...'
            if not part_brief:
                part_brief = '-'
            contents.append([part_brief, '', False, ''])
        elif part_type == 'TextBlock':
            contents.append(['TextBlock', padding, False, ''])
            part_ref = part_dict['textref']
            if len(part_ref)>40:
                part_ref = part_ref[:35] + '...'
            if not part_ref:
                part_ref = '-'
            contents.append([part_ref, '', False, ''])
        elif part_type == 'SectionPlaceHolder':
            section_name = part_dict['placename']
            if section_name:
                section_name = "Section " + section_name
            else:
                section_name = "Section -None-"
            if len(section_name)>40:
                section_name = section_name[:35] + '...'
            contents.append([section_name, padding, False, ''])
            part_brief = html.escape(part_dict.get('brief',''))
            if len(part_brief)>40:
                part_brief = part_brief[:35] + '...'
            if not part_brief:
                part_brief = '-'
            contents.append([part_brief, '', False, ''])
        elif part_type == 'Text':
            contents.append(['Text', padding, False, ''])
            # in this case part_dict is the text string rather than a dictionary
            if len(part_dict)<40:
                part_str = html.escape(part_dict)
            else:
                part_str = html.escape(part_dict[:35] + '...')
            if not part_str:
                part_str = '-'
            contents.append([part_str, '', False, ''])
        elif part_type == 'HTMLSymbol':
            contents.append(['Symbol', padding, False, ''])
            part_text = part_dict['text']
            if len(part_text)<40:
                part_str = html.escape(part_text)
            else:
                part_str = html.escape(part_text[:35] + '...')
            if not part_str:
                part_str = '-'
            contents.append([part_str, '', False, ''])
        elif part_type == 'Comment':
            contents.append(['Comment', padding, False, ''])
            part_text = part_dict['text']
            if len(part_text)<33:
                part_str =  "&lt;!--" + part_text + '--&gt;'
            else:
                part_str = "&lt;!--" + part_text[:31] + '...'
            if not part_str:
                part_str = '&lt;!----&gt;'
            contents.append([part_str, '', False, ''])
        elif part_type == 'ClosedPart':
            if 'attribs' in part_dict:
                tag_name = "&lt;%s ... /&gt;" % part_dict['tag_name']
            else:
                tag_name = "&lt;%s /&gt;" % part_dict['tag_name']
            contents.append([tag_name, padding, False, ''])
            part_brief = html.escape(part_dict.get('brief',''))
            if len(part_brief)>40:
                part_brief = part_brief[:35] + '...'
            if not part_brief:
                part_brief = '-'
            contents.append([part_brief, '', False, ''])
        elif part_type == 'Part':
            if 'attribs' in part_dict:
                tag_name = "&lt;%s ... &gt;" % part_dict['tag_name']
            else:
                tag_name = "&lt;%s&gt;" % part_dict['tag_name']
            contents.append([tag_name, padding, False, ''])
            part_brief = html.escape(part_dict.get('brief',''))
            if len(part_brief)>40:
                part_brief = part_brief[:35] + '...'
            if not part_brief:
                part_brief = '-'
            contents.append([part_brief, '', False, ''])
        else:
            contents.append(['UNKNOWN', padding, False, ''])
            contents.append(['ERROR', '', False, ''])

        # UP ARROW
        if rows == 2:
            # second line in table cannot move upwards
            contents.append(['', '', False, '' ])
        else:
            contents.append(['&uarr;', 'width : 1%;', True, part_location_string])

        # UP RIGHT ARROW
        if u_r_flag:
            contents.append(['&nearr;', 'width : 1%;', True, part_location_string])
        else:
            contents.append(['', '', False, '' ])

        # DOWN ARROW
        if (indent == 2) and (index == last_index):
            # the last line at this top indent has been added, no down arrow
            contents.append(['', '', False, '' ])
        else:
            contents.append(['&darr;', 'width : 1%;', True, part_location_string])

        # DOWN RIGHT ARROW
        # set to empty, when next line is created if down-right not applicable
        contents.append(['', '', False, '' ])

        # EDIT
        contents.append(['Edit', 'width : 1%;', True, part_location_string])

        # INSERT or APPEND
        if part_type == 'Part':
            contents.append(['Insert', 'width : 1%;text-align: center;', True, part_location_string])
        else:
            contents.append(['Append', 'width : 1%;text-align: center;', True, part_location_string])

        # COPY
        contents.append(['Copy', 'width : 1%;', True, part_location_string])

        # PASTE
        contents.append(['Paste', 'width : 1%;', True, part_location_string])

        # REMOVE
        contents.append(['Remove', 'width : 1%;', True, part_location_string])

        u_r_flag = False
        if part_type == 'Part':
            if last_row_at_this_level and (part_dict['tag_name'] != 'script') and (part_dict['tag_name'] != 'pre'):
                # add down right arrow in previous row at this level, get loc_string from adjacent edit cell
                editcell = contents[last_row_at_this_level *11-5]
                loc_string = editcell[3]
                contents[last_row_at_this_level *11-6] = ['&searr;', 'width : 1%;', True, loc_string]
            last_row_at_this_level = rows
            rows = domtree(part_dict, part_location_string, contents, part_string_list, rows, indent)
            # set u_r_flag for next item below this one
            if  (part_dict['tag_name'] != 'script') and (part_dict['tag_name'] != 'pre'):
                u_r_flag = True
        else:
            last_row_at_this_level =rows

    return rows





####### SKIPOLE WEB FRAMEWORK #######
#
# off_piste.py  - Specialist functions
#
# This file is part of the Skipole web framework
#
#
# Somewhat risky functions, hence 'off_piste'.
#

import copy, os, json, collections

from ..ski import skiboot, tag, widgets
from ..ski.excepts import ServerError

from . import project_loaded


def _get_proj(project):
    "Returns the Project with name project"
    project_loaded(project)
    return skiboot.getproject(project)


def _get_project_template_pages(project):
    "yields all template pages in the project"
    proj = _get_proj(project)
    identitems = proj.identitems
    for ident, item  in identitems.items():
        # item is taken direct from the project without copying or importing sections
        if item.page_type == 'TemplatePage':
            yield item


def set_widget_field_value(project, widget_name, widget_field_name, new_value, new_brief=None):
    """For each template page, with a given widget name and field name, set the field value, and widget brief.

       For every template page in the project, if it contains a widget with the given widget name and field name,
       then set the field value, and widget brief.
       The field can only be a single value field, not dictionary, table or list fields.
       If no brief is given, or an empty string is given, then the brief is not changed."""
    field_value = str(new_value)
    proj = _get_proj(project)
    for page in _get_project_template_pages(project):
        if widget_name in page.widgets:
            widget = page.widgets[widget_name]
            field_arg = widget.get_field_arg(widget_field_name)
            if field_arg is None:
                continue
            field_info = widget.field_arg_info(field_arg)
            # (field name, fieldvalue, str_fieldvalue, fieldarg class string, field type, field.valdt, field.jsonset, field.cssclass, field.csstyle)
            if field_info[3] != 'args':
                raise ServerError(message="Invalid field type in widget in page %s " % (page.ident,))
            if (field_info[4] == "ident" or field_info[4] == "url") and field_value.isdigit():
                # A digit is being input as a page ident
                field_value = project + '_' + field_value
            widget.set_field_value(field_arg, field_value)
            if new_brief:
                widget.brief = new_brief
            proj.save_page(page)


def insert_div_in_body(project, css_class, brief):
    """Inserts a div with the given css class as the containing div in each template page body

       For every template page in the project, inserts a div with the given css class attribute
       as the containing div in each page body (thus encapsulating all other body contents),
       unless a div with this class already exists at this point.
       Sets the containing div brief description with the brief given here."""
    proj = _get_proj(project)
    for page in _get_project_template_pages(project):
        body = page.body
        if (len(body) == 1) and hasattr(body[0], 'attribs') and body[0].get_attrib_value('class'):
            # body has one item with a class attribute
            if body[0].get_attrib_value('class') == css_class:
                # container div already exists
                if body[0].brief != brief:
                    body[0].brief = brief
                continue
        # so insert a div
        contents = copy.deepcopy(body.parts)
        newdiv = tag.Part(tag_name='div', attribs={'class':css_class}, brief=brief)
        newdiv.parts = contents
        body.parts = [newdiv]
        proj.save_page(page)


def set_backcol_in_pages(project, backcol):
    """Sets a background colour in the <html> tags of template pages.

       For every template page in the project, which has the option to set a background colour enabled,
       this function sets the background colour in the top <html> tag.
       The backcol argument should be a colour string, such as #FF0000."""
    proj = _get_proj(project)
    for page in _get_project_template_pages(project):
        page.backcol = backcol
        proj.save_page(page)


def set_bodyclass_in_pages(project, bodyclass):
    """For every template page in the project, sets the CSS class into the <body> tag."""
    proj = _get_proj(project)
    for page in _get_project_template_pages(project):
        page.body.set_class(bodyclass)
        proj.save_page(page)


def append_scriptlink_in_pages(project, label):
    """Appends a script link to the head of each page

       For every template page in the project, appends a script link
       to the head section pointing to the page with the given label.
       Generally used to set a link to a javascript file in all your pages."""
    proj = _get_proj(project)
    if "/" in label:
        scriptlink = tag.Part(tag_name = "script", attribs={"src":label}, brief='script link to %s' % (label,))
    else:
        scriptlink = tag.Part(tag_name = "script", attribs={"src":"{" + label + "}"}, brief='script link to %s' % (label,))
    for page in _get_project_template_pages(project):
        page.head.append(scriptlink)
        proj.save_page(page)


def set_widget_css_to_default(project):
    """For each template page, set all widget css class fields to their default values.

       For each template page in the project, set all widget CSS class fields to their
       default values, if a default has been set.
       Note: This does not change widgets in sections or svg pages."""
    proj = _get_proj(project)
    defaults = {}
    defaultsfile = skiboot.project_defaults(proj_ident=project)
    if os.path.isfile(defaultsfile):
        # load defaultsfile into defaults
        with open(defaultsfile, 'r') as fp:
            defaults = json.load(fp, object_pairs_hook=collections.OrderedDict)
    else:
        return
    if not defaults:
        return
    if 'widgets' not in defaults:
        return
    widget_defaults = defaults['widgets']
    # widget_defaults is a dictionary with keys of widget modules, and values which are themselves dictionaries
    # the value dictionaries have keys of the widget class, and values which are dictionaries of field:css class
    for page in _get_project_template_pages(project):
        change_flag = False
        for widget in page.widgets.values():
            widget_module = widget.__class__.__module__.split('.')[-1]
            widget_class = widget.__class__.__name__
            field_css = {}
            if (widget_module in widget_defaults) and (widget_class in widget_defaults[widget_module]):
                field_css = widget_defaults[widget_module][widget_class]
            else:
                continue
            if not field_css:
                continue                
            # for each css field in the widget, if it has a default in field_css, set it
            args = widget.field_arguments_single()
            for field in args:
                field_arg = field[0]
                field_type = field[2]
                if field_type != 'cssclass':
                    continue
                if (field_arg in field_css) and (field_css[field_arg]):
                    widget.set_field_value(field_arg, field_css[field_arg])
                    change_flag = True
        if change_flag:
            proj.save_page(page)







"""Functions for editing a responder"""

import inspect

from ..ski import skiboot, responders, dump_project
from ..ski.page_class_definition import RespondPage
from ..ski.excepts import ServerError

from . import get_proj_page

from .info_tuple import ResponderInfo



def _label_or_ident(ident):
    "Returns ident tuple or label"
    if isinstance(ident, skiboot.Ident):
        return ident.to_tuple()
    if ident is None:
        return ''
    return str(ident)


def list_responders():
    """Returns a list of lists, each inner list consisting of
       0) the responder class name
       1) the responder module name"""
    responderlist = []
    for name,responder in inspect.getmembers(responders, inspect.isclass):
        if issubclass(responder, responders.Respond):
            if name == 'Respond': continue
            responderlist.append([name, responder.module_name()])
    return responderlist


def responder_info(project, pagenumber, pchange=None):
    """Return a ResponderInfo named tuple"""
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "RespondPage":
        raise ServerError(message = "Invalid page type")
    responder = page.responder

    if responder.widgfield_required and responder.widgfield:
        widgfield = responder.widgfield.to_str_tuple()
    else:
        widgfield = None

    if responder.alternate_ident_required and responder.alternate_ident:
        alternate_ident = responder.alternate_ident
    else:
        alternate_ident = None

    if responder.target_ident_required and responder.target_ident:
        target_ident = responder.target_ident
    else:
        target_ident = None

    if responder.allowed_callers_required and responder.allowed_callers:
        allowed_callers = [_label_or_ident(ident) for ident in responder.allowed_callers]
    else:
        allowed_callers = []

    if responder.validate_option_available:
        validate_option = responder.validate_option
    else:
        validate_option = False

    if responder.validate_option_available and responder.validate_fail_ident:
        validate_fail_ident = responder.validate_fail_ident
    else:
        validate_fail_ident = None

    # submit option

    if responder.submit_option_available:
        submit_option = responder.submit_option
    else:
        submit_option = False

    if responder.submit_required or submit_option:
        if responder.submit_list:
            submit_list = responder.submit_list[:]
        else:
            submit_list = []
        if responder.fail_ident:
            fail_ident = responder.fail_ident
        else:
            fail_ident = None
    else:
        submit_list = []
        fail_ident = None

    # fields

    field_options = responder.field_options.copy()
    field_values_list = []
    field_list = []
    single_field_value = []
    single_field = None

    if field_options['fields']:

        # field_values_list
        if field_options['field_values'] and ( not field_options['single_field'] ):
            field_vals = responder.fields
            for field, value in field_vals.items():
                if isinstance(field, skiboot.WidgField):
                    f = field.to_tuple()
                elif isinstance(field, skiboot.Ident):
                    f = str(field)
                else:
                    f = field
                if isinstance(value, skiboot.WidgField):
                    v = value.to_tuple()
                elif isinstance(value, skiboot.Ident):
                    v = str(value)
                else:
                    v = value
                field_values_list.append([f,v])

        # field_list
        if (not field_options['field_values']) and (not field_options['single_field']):
            field_vals = responder.fields
            for field in field_vals:
                if isinstance(field, skiboot.WidgField):
                    f = field.to_tuple()
                elif isinstance(field, skiboot.Ident):
                    f = str(field)
                else:
                    f = field
                field_list.append(f)

        if field_options['single_field']:

            # single_field_value - still to do
            # as there is currently no responder which takes a single field and value
            #????????????????????????????????????????????????????????

            # single_field
            if not field_options['field_values']:
                if not responder.responder_fields.keys():
                    single_field = ''
                else:
                    single_field = list(responder.responder_fields.keys())[0]

    return ResponderInfo(responder.__class__.__name__,
                         responder.module_name(),
                         responder.widgfield_required,
                         widgfield,
                         responder.alternate_ident_required,
                         _label_or_ident(alternate_ident),
                         responder.target_ident_required,
                         _label_or_ident(target_ident),
                         responder.allowed_callers_required,
                         allowed_callers,
                         responder.validate_option_available,
                         validate_option,
                         _label_or_ident(validate_fail_ident),
                         responder.submit_option_available,
                         submit_option,
                         responder.submit_required,
                         submit_list,
                         _label_or_ident(fail_ident),
                         field_options,
                         field_values_list,
                         field_list,
                         single_field_value,
                         single_field
                       )


def get_submit_list(project, pagenumber, pchange):
    "Returns a copy of the submit list"
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "RespondPage":
        raise ServerError(message = "Invalid page type")
    responder = page.responder
    if responder.submit_required or responder.submit_option:
        if responder.submit_list:
            return responder.submit_list[:]
        else:
            return []
    raise ServerError(message = "Invalid responder - no submit list available")


def set_submit_list(project, pagenumber, pchange, submit_list):
    "Sets the submit list"
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "RespondPage":
        raise ServerError(message = "Invalid page type")
    responder = page.responder
    if (not responder.submit_required) and (not responder.submit_option):
        raise ServerError(message = "Invalid responder - no submit list available")
    responder.submit_list = submit_list[:]
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def set_widgfield(project, pagenumber, pchange, widgfield):
    "sets responder widgfield, returns new pchange"
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "RespondPage":
        raise ServerError(message = "Invalid page type")
    responder = page.responder
    if not responder.widgfield_required:
        raise ServerError(message="Invalid submission, this responder does not have a widgfield")
    responder.widgfield = skiboot.make_widgfield(widgfield)._replace(i='')
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def set_alternate_ident(project, pagenumber, pchange, ident):
    "Sets the alternate page"
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "RespondPage":
        raise ServerError(message = "Invalid page type")
    responder = page.responder
    if not responder.alternate_ident_required:
        raise ServerError(message="Invalid submission, this responder does not have an alternate page")
    a_i = skiboot.make_ident_or_label_or_url(ident, proj_ident=project)
    if a_i is None:
        raise ServerError(message="Invalid alternate ident")
    responder.alternate_ident = a_i
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def set_target_ident(project, pagenumber, pchange, ident):
    "Sets the target page"
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "RespondPage":
        raise ServerError(message = "Invalid page type")
    responder = page.responder
    if not responder.target_ident_required:
        raise ServerError(message="Invalid submission, this responder does not have a target ident")
    t_i = skiboot.make_ident_or_label_or_url(ident, proj_ident=project)
    if t_i is None:
        raise ServerError(message="Invalid target ident")
    responder.target_ident = t_i
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def set_validate_fail_ident(project, pagenumber, pchange, ident):
    "Sets the fail page for validate error"
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "RespondPage":
        raise ServerError(message = "Invalid page type")
    responder = page.responder
    if not responder.validate_option_available:
        raise ServerError(message="Invalid submission, this responder does not have a validate option")
    v_f_i = skiboot.make_ident_or_label_or_url(ident, proj_ident=project)
    if v_f_i is None:
        raise ServerError(message="Invalid validate fail ident")
    responder.validate_fail_ident = v_f_i
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def set_fail_ident(project, pagenumber, pchange, ident):
    "Sets the fail page"
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "RespondPage":
        raise ServerError(message = "Invalid page type")
    responder = page.responder
    if not (responder.submit_required or responder.submit_option_available):
        raise ServerError(message="Invalid submission, this responder does not have a fail ident")
    f_i = skiboot.make_ident_or_label_or_url(ident, proj_ident=project)
    if f_i is None:
        raise ServerError(message="Invalid fail ident")
    responder.fail_ident = f_i
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def add_allowed_caller(project, pagenumber, pchange, ident):
    "Add an allowed caller"
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "RespondPage":
        raise ServerError(message = "Invalid page type")
    responder = page.responder
    if not responder.allowed_callers_required:
        raise ServerError(message="Invalid submission, this responder does not have allowed callers")
    a_c = skiboot.make_ident_or_label(ident, proj_ident=project)
    if a_c is None:
        raise ServerError(message="Invalid allowed caller")
    # check not already in list
    allowed_callers = [str(idt) for idt in responder.allowed_callers ]
    s_a_c = str(a_c)
    if s_a_c in allowed_callers:
        raise ServerError(message='This allowed caller already exists')
    responder.allowed_callers.append(a_c)
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def delete_allowed_caller(project, pagenumber, pchange, ident):
    "Deletes an allowed caller"
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "RespondPage":
        raise ServerError(message = "Invalid page type")
    responder = page.responder
    if not responder.allowed_callers_required:
        raise ServerError(message="Invalid submission, this responder does not have allowed callers")
    d_a_c = skiboot.make_ident_or_label(ident, proj_ident=project)
    if d_a_c is None:
        raise ServerError(message="Invalid caller to delete")
    allowed_callers = [str(idt) for idt in responder.allowed_callers ]
    s_d_a_c = str(d_a_c)
    try:
        idx = allowed_callers.index(s_d_a_c)
    except ValueError:
        raise ServerError('This allowed caller does not exist')
    del responder.allowed_callers[idx]
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def remove_field(project, pagenumber, pchange, field):
    "Deletes a field"
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "RespondPage":
        raise ServerError(message = "Invalid page type")
    responder = page.responder
    # field options
    f_options = responder.field_options
    if not f_options['fields']:
        raise ServerError(message="Invalid submission, this responder does not have fields")
    if f_options['widgfields']:
        # ensure the field to remove is a widgfield
        field = skiboot.make_widgfield(field)
    if field in responder.fields:
        del responder.fields[field]
    else:
        raise ServerError(message="Field not found")
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def add_field_value(project, pagenumber, pchange, field, value):
    "Adds a field and value"
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "RespondPage":
        raise ServerError(message = "Invalid page type")
    responder = page.responder
    # field options
    f_options = responder.field_options
    if not f_options['fields']:
        raise ServerError(message="Invalid submission, this responder does not have fields")
    if not f_options['field_values']:
        raise ServerError(message="Invalid submission, this responder does not have values")
    if (not value) and (not f_options['empty_values_allowed']):
        raise ServerError(message="Invalid submission, empty field values are not allowed")
    responder.set_field(field, value)
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def add_field(project, pagenumber, pchange, field):
    "Adds a field, without a value, options['field_values'] must be False"
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "RespondPage":
        raise ServerError(message = "Invalid page type")
    responder = page.responder
    # field options
    f_options = responder.field_options
    if not f_options['fields']:
        raise ServerError(message="Invalid submission, this responder does not have fields")
    if f_options['field_values']:
        raise ServerError(message="Invalid submission, this responder requires fields with values")
    responder.set_field(field, '')
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def set_single_field(project, pagenumber, pchange, field):
    "Sets a single field, without a value, options['field_values'] must be False and options['single_field'] must be True"
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "RespondPage":
        raise ServerError(message = "Invalid page type")
    responder = page.responder
    # field options
    f_options = responder.field_options
    if not f_options['fields']:
        raise ServerError(message="Invalid submission, this responder does not have fields")
    if f_options['field_values']:
        raise ServerError(message="Invalid submission, this responder requires fields with values")
    if not f_options['single_field']:
        raise ServerError(message="Invalid submission, this function can only be used with single_field responders")
    responder.set_fields({ field:'' })
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def toggle_validate_option(project, pagenumber, pchange):
    "Toggles the validate option, returns (pchange, validate_option)"
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "RespondPage":
        raise ServerError(message = "Invalid page type")
    responder = page.responder
    if not responder.validate_option_available:
        raise ServerError(message="Invalid submission, this responder does not accept the validate option")
        
    if responder.validate_option:
        responder.validate_option = False
    else:
        responder.validate_option = True
    # save the altered page, and return the page.change uuid and validate option
    return proj.save_page(page), responder.validate_option


def toggle_submit_option(project, pagenumber, pchange):
    "Toggles the submit option, returns (pchange, submit_option)"
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "RespondPage":
        raise ServerError(message = "Invalid page type")
    responder = page.responder
    if not responder.submit_option_available:
        raise ServerError(message="Invalid submission, this responder does not accept the submit option")
        
    if responder.submit_option:
        responder.submit_option = False
        responder.submit_list = []
    else:
        responder.submit_option = True
    # save the altered page, and return the page.change uuid and submit option
    return proj.save_page(page), responder.submit_option


def all_responders(project):
    "Returns a list of all responder idents in the project"
    # Get the root folder dictionary
    responder_idents = []
    rootdict = dump_project.folder_to_OD(project, skiboot.Ident(project, 0))
    pages = _all_pages(rootdict)
    for page_dict in pages:
        if 'RespondPage' in page_dict:
            responder_idents.append(page_dict["ident"])
    responder_idents.sort()
    return responder_idents 


def _all_pages(topfolder):
    """Return list of all page dictionaries in a folder and its sub folders
       Where topfolder is the folder dictionary"""
    # search through folders
    page_list = []
    if "folders" in topfolder:
        for folder in topfolder['folders'].values():
            page_list.extend(_all_pages(folder))
    if "pages" in topfolder:
        for page in topfolder['pages'].values():
            page_list.append(page)
    return page_list










"""Contains widgets for textarea forms"""


from .. import skiboot, tag, excepts
from . import Widget, ClosedWidget, FieldArg, FieldArgList, FieldArgTable, FieldArgDict

class SubmitTextArea(Widget):
    """Defines a form containing a div with a text area input field"""

    error_location = (0,0,0)

    arg_descriptions = {'error_class':FieldArg("cssclass", ""),
                        'action':FieldArg("url", ''),
                        'hidden_field1':FieldArg("text", '', valdt=True),
                        'hidden_field2':FieldArg("text", '', valdt=True),
                        'hidden_field3':FieldArg("text", '', valdt=True),
                        'hidden_field4':FieldArg("text", '', valdt=True),
                        'button_text':FieldArg("text", 'Submit'),
                        'button_class':FieldArg("cssclass", ''),
                        'inputdiv_class':FieldArg("cssclass", ''),
                        'input_text':FieldArg("text", '', valdt=True, jsonset=True),
                        'textarea_class':FieldArg("cssclass", ''),
                        'textarea_style':FieldArg("cssstyle", ''),
                        'rows':FieldArg("text", '4'),
                        'cols':FieldArg("text", '40')
                       }


    def __init__(self, name=None, brief='', **field_args):
        """
        action: The page ident, label, url this button links to
        hidden_field1: A hidden field value, leave blank if unused, name used as the get field name
        hidden_field2: A second hidden field value, leave blank if unused, name used as the get field name
        hidden_field3: A third hidden field value, leave blank if unused, name used as the get field name
        hidden_field4: A fourth hidden field value, leave blank if unused, name used as the get field name
        button_text: The text on the button
        button_class: class set on the buttons
        inputdiv_class: the class attribute of the div which contains the input text area and button
        error_class: The class applied to the paragraph containing the error message on error.
        input_text: The default text in the text area, field name used as the name attribute
        textarea_class: The class applied to the textarea
        textarea_style: The style applied to the textarea
        rows:  The number of rows of the text area
        cols: The number of columns of the text area
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        # hidden error at div 0
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        # The form
        self[1] = tag.Part(tag_name='form', attribs={"role":"form", "method":"post"})
        # div containing textarea and button
        self[1][0] = tag.Part(tag_name='div')
        # the text input field
        self[1][0][0] = tag.Part(tag_name="textarea")
        # the submit button
        self[1][0][1] = tag.ClosedPart(tag_name="input")


    def _build(self, page, ident_list, environ, call_data, lang):
        "build the form"
        if self.get_field_value('error_class'):
            self[0].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[0].del_one_attrib("style")
        if not self.get_field_value("action"):
            # setting self._error replaces the entire tag
            self._error = "Warning: No form action"
            return
        actionurl = skiboot.get_url(self.get_field_value("action"),  proj_ident=page.proj_ident)
        if not actionurl:
            # setting self._error replaces the entire tag
            self._error = "Warning: broken link"
            return
        # update the action of the form
        self[1].update_attribs({"action": actionurl})
        # the class of the div holding textarea and button
        if self.get_field_value('inputdiv_class'):
            self[1][0].attribs = {"class": self.get_field_value('inputdiv_class')}

        # set up the text area
        self[1][0][0].attribs = {"name":self.get_formname('input_text')}

        if self.get_field_value('rows'):
            self[1][0][0].update_attribs({"rows": self.get_field_value('rows')})

        if self.get_field_value('cols'):
            self[1][0][0].update_attribs({"cols": self.get_field_value('cols')})

        if self.get_field_value('textarea_class'):
            self[1][0][0].update_attribs({"class": self.get_field_value('textarea_class')})

        if self.get_field_value("textarea_style"):
            self[1][0][0].update_attribs({"style":self.get_field_value("textarea_style")})

        if self.get_field_value('input_text'):
            self[1][0][0][0] = self.get_field_value('input_text')

        # set up the submit button
        if self.get_field_value('button_text'):
            submit = self.get_field_value('button_text')
        else:
            submit = "Submit"

        if self.get_field_value('button_class'):
            self[1][0][1].attribs = {"value":submit, "type":"submit", "class":self.get_field_value('button_class')}
        else:
            self[1][0][1].attribs = {"value":submit, "type":"submit"}

        # add ident and four hidden fields
        self.add_hiddens(self[1], page)


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div> <!-- with widget id and class widget_class -->
  <div> <!-- normally hidden paragraph, with class error_class -->
    <p> <!-- Any error text appears here --> </p>
  </div>
  <form method="post">  <!-- with action given by action field -->
    <div> <!-- the class attribute set to inputdiv_class -->
      <textarea> <!-- class attribute set to textarea_class. Rows, cols set to the given rows and columns -->
        <!-- textarea content set to input_text -->
      </textarea>
      <input type="submit" /> <!-- button value set to button_text -->
    </div>
    <!-- hidden input fields -->                              
  </form>
</div>"""



class TextArea1(Widget):
    """Defines an input field only (no form, label or submit button)
       does not display errors"""

    # This class does not display any error messages
    display_errors = False

    # js_validators is a class attribute, True if javascript validation is enabled
    js_validators=True

    arg_descriptions = {
                        'input_text':FieldArg("text", '', valdt=True, jsonset=True),
                        'rows':FieldArg("text", '4'),
                        'cols':FieldArg("text", '40')
                       }
    def __init__(self, name=None, brief='', **field_args):
        """
        input_text: The default text in the text input field, field name used as name attribute
        rows:  The number of rows of the text area
        cols: The number of columns of the text area
        """
        Widget.__init__(self, name=name, tag_name="textarea", brief=brief, **field_args)
        self[0] = ''

    def _build(self, page, ident_list, environ, call_data, lang):
        "build the input field"
        self.update_attribs({"name":self.get_formname('input_text')})
        if self.get_field_value('input_text'):
            self[0] = self.get_field_value('input_text')
        if self.get_field_value('rows'):
            self.update_attribs({"rows": self.get_field_value('rows')})
        if self.get_field_value('cols'):
            self.update_attribs({"cols": self.get_field_value('cols')})
 
    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
    <textarea> <!-- with widget id and class widget_class. Rows, cols set to the given rows and columns -->
      <!-- textarea content set to input_text -->
    </textarea>
"""



class TextArea2(Widget):
    """Defines a div containing a hidden error paragraph
       with a textarea input field (no form or submit button)
       A label to the left of the textarea can have a class and text set"""

    # js_validators is a class attribute, True if javascript validation is enabled
    js_validators=True

    error_location = (0,0,0)

    arg_descriptions = {
                        'label':FieldArg("text", 'Your input:'),
                        'label_class':FieldArg("cssclass", ''),
                        'label_style':FieldArg("cssstyle", ''),
                        'error_class':FieldArg("cssclass", ''),
                        'input_text':FieldArg("text", '', valdt=True, jsonset=True),
                        'textarea_class':FieldArg("cssclass", ''),
                        'textarea_style':FieldArg("cssstyle", ''),
                        'redstar':FieldArg("boolean", False),
                        'redstar_class':FieldArg("cssclass", ''),
                        'redstar_style':FieldArg("cssstyle", ''),
                        'rows':FieldArg("text", '4'),
                        'cols':FieldArg("text", '40')
                       }
    def __init__(self, name=None, brief='', **field_args):
        """
        label: The text appearing in a label tag to the left of the input field
        label_class: The css class of the label to the left of the input field
        error_class: The class applied to the hidden paragraph on error.
        input_text: The default text in the text input field, field name used as name attribute
        redstar: If True a red asterisk is shown by the side of the input field
        textarea_class: The class applied to the textarea
        textarea_style: The style applied to the textarea
        rows:  The number of rows of the text area
        cols: The number of columns of the text area
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        # hidden error
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        # the label
        self[1] = tag.Part(tag_name="label", hide_if_empty=True)
        # the textarea input field
        self[2] = tag.Part(tag_name="textarea")
        self[2][0] =''


    def _build(self, page, ident_list, environ, call_data, lang):
        "build the input field"
        if self.get_field_value('error_class'):
            self[0].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[0].del_one_attrib("style")
        if self.get_field_value('label_class'):
            self[1].update_attribs({"class": self.get_field_value('label_class')})
        if self.get_field_value('label_style'):
            self[1].update_attribs({"style": self.get_field_value('label_style')})
        if self.get_field_value("label"):
            self[1][0] = self.get_field_value("label")
        # set an id in the textarea for the 'label for' tag
        self[2].insert_id()
        self[1].update_attribs({'for':self[2].get_id()})
        self[2].update_attribs({"name":self.get_formname('input_text')})
        if self.get_field_value('textarea_class'):
            self[2].update_attribs({"class": self.get_field_value('textarea_class')})
        if self.get_field_value("textarea_style"):
            self[2].update_attribs({"style":self.get_field_value("textarea_style")})

        if self.get_field_value('rows'):
            self[2].update_attribs({"rows": self.get_field_value('rows')})
        if self.get_field_value('cols'):
            self[2].update_attribs({"cols": self.get_field_value('cols')})
        if self.get_field_value('input_text'):
            self[2][0] = self.get_field_value('input_text')
        # redstar
        if self.get_field_value('redstar'):
            self[3] = tag.Part(tag_name="span")
            if self.get_field_value('redstar_style'):
                self[3].update_attribs({"style":self.get_field_value('redstar_style')})
            if self.get_field_value('redstar_class'):
                self[3].update_attribs({"class":self.get_field_value('redstar_class')})
            self[3][0] = '*'




    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
  <div> <!-- normally hidden paragraph, with class error_class -->
    <p> <!-- Any error text appears here --> </p>
  </div>
  <label> <!-- with class set to label_class -->
          <!-- content set to label -->
  </label>
  <textarea> <!-- class attribute set to textarea_class. Rows, cols set to the given rows and columns -->
    <!-- textarea content set to input_text -->
  </textarea>
  <span>*</span> <!-- shown if redstar is True -->
</div>"""





"""
This module defines the WSGIApplication class and SkiCall class
"""


import copy, os, cgi, collections, html, pprint, json, shutil, uuid, sys, traceback, re

from http import cookies

# a search for anything none-alphanumeric and not an underscore
_AN = re.compile('[^\w]')

from . import skiboot, read_json
from .excepts import ValidateError, ServerError, FailPage, ErrorMessage, GoTo, PageError
from .. import textblocks


# ServerError raised in this module use codes 9000 to 9100


class WSGIApplication(object):
    """The WSGIApplication - an instance being a callable WSGI application"""

    def __init__(self, project, projectfiles=None, proj_data={}, start_call=None, submit_data=None, end_call=None, url="/"):
        """Initiates a Project instance"""
        if _AN.search(project):
            raise ServerError(message="Error: Invalid project name, alphanumeric only")
        if '_' in project:
            raise ServerError(message="Error: Invalid project name, alphanumeric only (no underscore).")
        self._proj_ident = project
        self.projectfiles = projectfiles
        self.proj_data = proj_data
        self.start_call = start_call
        self.submit_data = submit_data
        self.end_call = end_call

        # initially, sub projects can be added
        self.rootproject = True

        self.brief = "Project %s" % project
        self.version = "0.0.0"
        # The url of the root folder
        url=url.strip("/").lower()
        if url:
            self.url = "/" + url + "/"
        else:
            self.url = "/"
        # The root Folder
        self.root = None
        # dictionary of special pages, key = label: value = page ident
        self.special_pages = {}
        # dictionary of sections, key = name: value = section
        self.sections = {}

        # an ordered dictionary of {proj_ident: url,...}, ordered by length of url
        self._subproject_paths = collections.OrderedDict()
        # self.subprojects is a dictionary of sub projects {proj_ident: Project instance,.....}
        self.subprojects = {}

        # dictionary of idents: to folder or page, apart from root
        # note keys are full Ident instances, values are  folder or page instances
        self.identitems = {}

        # Create an instance of the AccessTextBlocks class for this project
        self.textblocks = textblocks.AccessTextBlocks(self._proj_ident, projectfiles, skiboot.default_language())

        # maintain a cach dictionary of paths against idents {path:ident}
        self._paths = {}

        # load project from json files
        self.load_from_json()

        # and add this project to the project register
        skiboot.add_to_project_register(self)


    def __call__(self, environ, start_response):
        "Defines this projects callable as the wsgi application"
        status, headers, data = self.respond(environ)
        start_response(status, headers)
        return data

    def set_default_language(self, language):
        "Sets the project default language"
        self.textblocks.default_language = language

    def get_default_language(self):
        "Returns the project default language"
        return self.textblocks.default_language

    default_language = property(get_default_language, set_default_language)

    def clear_cache(self):
        "clear the cache of paths"
        self._paths = {}


    def load_from_json(self):
        "Loads project with data saved in project.json file"
        projectdict = read_json.create_project(self._proj_ident, self.projectfiles)
        self.default_language = projectdict['default_language']
        self.brief = projectdict['brief']
        self.version = projectdict['version']
        self.special_pages = projectdict['specialpages']
        self.sections = projectdict['sections']
        self.root = projectdict['siteroot']
        itemlist = projectdict['itemlist']
        self.identitems = {}
        if itemlist:
            for item in itemlist:
                self.identitems[item.ident] = item


    @property
    def max_ident_num(self):
        "Returns the maximum identnumber currently in use"
        maxnum = 0
        for ident in self.identitems:
            if ident.num > maxnum:
                    maxnum = ident.num
        return maxnum

    def list_section_names(self):
        "Returns a list of section names, alphabetacily ordered"
        if not self.sections:
            return []
        s = [ name for name in self.sections ]
        s.sort()
        return s

    def section(self, section_name, makecopy=True):
        "Returns a section, or a deep copy of a section, or None if the section name is not found"
        if section_name not in self.sections:
            return
        section = self.sections[section_name]
        if not makecopy:
            return section
        if section is None:
            return None
        return copy.deepcopy(section)
 

    def add_section(self, name, section):
        "Adds a section to the project, returns section.change uuid"
        # and save the section
        section.widgets = {}
        section.section_places = {}  # currently unused
        embedded = (name, '', None)
        section.set_idents(name, section.widgets, section.section_places, embedded)
        # set validators in section
        section.load_validator_scriptlinks()
        # set the section change number
        section.change = uuid.uuid4().hex
        self.sections[name] = section
        return section.change


    def delete_section(self, name):
        "Deletes a section"
        if name in self.sections:
            del self.sections[name]

    @property
    def ident_numbers(self):
        "return a list of ident numbers"
        num_list = [ ident.num for ident in self.identitems ]
        # insert the root
        num_list.insert(0,0)
        num_list.sort()
        return num_list

    def __getitem__(self, ident):
        """given an Ident, or a string version of ident, return page or folder. 
              If folder or respond page return the item, any other page, return a deep copy
              of the item. If item not found, return None"""
        ident = skiboot.Ident.to_ident(ident, proj_ident=self._proj_ident)
        if ident is None:
            return
        if ident.proj != self._proj_ident:
            return
        if ident.num == 0:
            return self.root
        if ident not in self.identitems:
            return
        item = self.identitems[ident]
        if item is None:
            return
        if item.page_type == 'Folder':
            return item
        if item.page_type == 'RespondPage':
            return item
        return copy.deepcopy(item)

    def add_item(self, parent_ident, item, ident=None):
        """Adds a new page or folder to the project, returns the item ident"""

        # check ident
        if ident is None:
            ident = skiboot.Ident(self._proj_ident, self.max_ident_num+1)
        else:
            ident = skiboot.make_ident(ident, self._proj_ident)
            if ident is None:
                raise ServerError(message="Sorry. Invalid ident", code=9000)
            if ident.num == 0:
                # cannot add the root folder
                raise ServerError(message="Sorry. Unable to add a new root", code=9001)
            if ident.proj != self._proj_ident:
                # must be in this project
                raise ServerError(message="Sorry. Invalid ident", code=9002)
            if ident in self.identitems:
                # ident must not exist
                raise ServerError(message="Sorry. The given ident already exists", code=9003)

        # check parent folder
        if parent_ident.proj != self._proj_ident:
            raise ServerError(message="Invalid parent ident", code=9004)
        parent = self.get_item(parent_ident)
        if parent is None:
            raise ServerError(message="Parent folder not found: Error in add_item method of Project class.", code=9005)
        if parent.page_type != 'Folder':
            raise ServerError(message="Parent not a folder", code=9006)

        if item.name in parent.pages:
            raise ServerError(message="Sorry, a page with that name already exists in the parent folder", code=9007)
        if item.name in parent.folders:
            raise ServerError(message="Sorry, a folder with that name already exists in the parent folder", code=9008)

        # set the item ident
        item.ident = ident

        # set this item name into the parent
        if item.page_type == 'Folder':
            if parent.restricted:
                item.set_restricted()
            parent.folders[item.name] = item.ident
        # if the page is a template or svg page, then set its idents
        # and store its widgets in the page's widgets dictionary and sectionplaceholders
        # in the page's sections directory
        elif (item.page_type == 'TemplatePage') or (item.page_type == 'SVG'):
            item.set_idents()
            parent.pages[item.name] = item.ident
            # now set validator modules in page
            if item.page_type == 'TemplatePage':
                item.load_validator_scriptlinks()
        else:
            parent.pages[item.name] = item.ident

        # set the parent change value
        parent.change = uuid.uuid4().hex
        item.parentfolder = parent

        # and finally, add the item
        self.identitems[item.ident] = item
        self.clear_cache()
        return item.ident


    def delete_item(self, itemident):
        """Deletes the page or folder with the given ident from the database."""
        #############  TO DO - if folder has contents, recursivly delete contents ########
        if itemident.num == 0:
            # cannot delete the root folder
            raise ServerError(message="Cannot delete the root folder", code=9009)
        if itemident.proj != self._proj_ident:
            # Must belong to this project
            raise ServerError(message="Cannot delete this item (does not belong to this project)", code=9010)
        if itemident not in self.identitems:
            raise ServerError(message="Item not found", code=9011)
        # get the item
        item = self.identitems[itemident]
        # get the items parent folder
        parentfolder = item.parentfolder
        if item.name in parentfolder.pages:
            del parentfolder.pages[item.name]
        if item.name in parentfolder.folders:
            del parentfolder.folders[item.name]
        parentfolder.change = uuid.uuid4().hex
        # del the item
        del self.identitems[itemident]
        self.clear_cache()


    def save_item(self, item, new_parent_ident=None):
        """Saves the page or folder - used to save an altered item, not to add a new one
           If new_parent_ident is not None, indicates the item has moved to a different folder
           returns the items new change uuid"""
        if item.page_type == 'Folder':
            return self.save_folder(item, new_parent_ident)
        else:
            return self.save_page(item, new_parent_ident)


    def save_page(self, item, new_parent_ident=None):
        """Saves the page - used to save an altered page, not to add a new one
           If new_parent_ident is not None, indicates the page has moved to a different folder
           Returns the new page.change uuid"""
        if item.page_type == 'Folder':
            raise ServerError(message="Invalid item, not a page.", code=9012)
        item_ident = item.ident
        if item_ident is None:
            raise ServerError(message="Unable to save page - no ident set", code=9013)
        if self._proj_ident != item_ident.proj:
            raise ServerError(message="Unable to save page - invalid ident", code=9014)
        if item_ident not in self.identitems:
            raise ServerError(message="This page ident does not exist", code=9015)
        old_parent = self.identitems[item_ident].parentfolder
        old_name = self.identitems[item_ident].name
        if new_parent_ident is not None:
            # So its a parent folder change
            if new_parent_ident.num is 0:
                # new parent is root
                new_parent = self.root
            else:
                new_parent = self.identitems[new_parent_ident]
            if new_parent == old_parent:
                new_parent_ident = None
        if (item.page_type == 'TemplatePage') or (item.page_type == 'SVG'):
            item.set_idents()
        # now set validator modules in page
        if item.page_type == 'TemplatePage':
            item.load_validator_scriptlinks()
        item.change = uuid.uuid4().hex
        if (old_name == item.name) and (new_parent_ident is None):
            # no folder change
            self.identitems[item_ident] = item
            self.clear_cache()
            return item.change
        if new_parent_ident is None:
            # so just a name change
            if item.name in old_parent.pages:
                raise ServerError(message="Sorry, a page with that name already exists", code=9016)
            if item.name in old_parent.folders:
                raise ServerError(message="Sorry, a folder with that name already exists", code=9017)
            if old_name in old_parent.pages:
                del old_parent.pages[old_name]
            old_parent.pages[item.name] = item_ident
            old_parent.change = uuid.uuid4().hex
            self.identitems[item_ident] = item
            self.clear_cache()
            return item.change
        # change of folder
        if item.name in new_parent.pages:
            raise ServerError(message="Sorry, a page with that name already exists", code=9018)
        if item.name in new_parent.folders:
            raise ServerError(message="Sorry, a folder with that name already exists", code=9019)
        if old_name in old_parent.pages:
            del old_parent.pages[old_name]
            old_parent.change = uuid.uuid4().hex
        new_parent.pages[item.name] = item_ident
        new_parent.change = uuid.uuid4().hex
        item.parentfolder = new_parent
        self.identitems[item_ident] = item
        self.clear_cache()
        return item.change


    def save_folder(self, item, new_parent_ident=None):
        """Saves the folder - used to save an altered folder, not to add a new one
           If new_parent_ident is not None, indicates the folder has moved to a different parent folder
           Returns the new folder.change uuid"""
        if item.page_type != 'Folder':
            raise ServerError(message="Invalid item, not a folder.", code=9020)
        item_ident = item.ident
        if item_ident is None:
            raise ServerError(message="Unable to save folder - no ident set", code=9021)
        if self._proj_ident != item_ident.proj:
            raise ServerError(message="Unable to save folder - invalid ident", code=9022)
        if item_ident.num == 0:
            if new_parent_ident:
                raise ServerError(message="Root folder cannot have new parent", code=9023)
            item.change = uuid.uuid4().hex
            self.root = item
            self.clear_cache()
            return item.change
        if item_ident not in self.identitems:
            raise ServerError(message="This folder ident does not exist", code=9024)
        old_parent = self.identitems[item_ident].parentfolder
        old_name = self.identitems[item_ident].name
        if new_parent_ident is not None:
            # So its a parent folder change
            if new_parent_ident.num is 0:
                # new parent is root
                new_parent = self.root
            else:
                new_parent = self.identitems[new_parent_ident]
            if new_parent == old_parent:
                new_parent_ident = None
        item.change = uuid.uuid4().hex
        if (old_name == item.name) and (new_parent_ident is None):
            # no parent folder change
            self.identitems[item_ident] = item
            self.clear_cache()
            return item.change
        if new_parent_ident is None:
            # so just a name change
            if item.name in old_parent.pages:
                raise ServerError(message="Sorry, a page with that name already exists", code=9025)
            if item.name in old_parent.folders:
                raise ServerError(message="Sorry, a folder with that name already exists", code=9026)
            if old_name in old_parent.folders:
                del old_parent.folders[old_name]
            old_parent.folders[item.name] = item_ident
            old_parent.change = uuid.uuid4().hex
            self.identitems[item_ident] = item
            self.clear_cache()
            return item.change
        # change of folder
        # A folder cannot be moved into a sub folder of itself

        folder_list = new_parent.parent_list()
        # folder list is a list of (name, identnumber) starting at root
        folder_ident_numbers = [ identnumber for name,identnumber in folder_list ]
        if item.ident.num in folder_ident_numbers:
            # item is a parent of new_parent
            raise ServerError(message="Sorry, a folder cannot be moved into a subfolder of itself.", code=9027)
        if item.name in new_parent.pages:
            raise ServerError(message="Sorry, a page with that name already exists", code=9028)
        if item.name in new_parent.folders:
            raise ServerError(message="Sorry, a folder with that name already exists", code=9029)
        if old_name in old_parent.folders:
            del old_parent.folders[old_name]
            old_parent.change = uuid.uuid4().hex
        new_parent.folders[item.name] = item_ident
        new_parent.change = uuid.uuid4().hex
        item.parentfolder = new_parent
        self.identitems[item_ident] = item
        self.clear_cache()
        return item.change



    def get_item(self, ident):
        """given an ident (Ident object or integer), return a
           folder or page from the database, if not found, return None.
           Note: the item is returned without copying and without sections imported"""
        if isinstance(ident, int):
            ident = skiboot.Ident(self._proj_ident, ident)
        elif ident.proj != self._proj_ident:
            return None
        if ident.num == 0:
            return self.root
        if ident not in self.identitems:
            return None
        return self.identitems[ident]


    def __iter__(self):
        "This iterator does not return the root folder"
        for ident in self.identitems:
            yield ident

    def __contains__(self, item):
        "Checks if this project contains folder, page or ident"
        if hasattr(item, 'ident'):
            ident = item.ident
        else:
            ident = skiboot.Ident.to_ident(item, self._proj_ident)
        if ident in self.identitems:
            return True
        if (ident.proj == self._proj_ident) and (ident.num is 0):
            return True
        return False

    def __len__(self):
        "This length does not include the root"
        return len(self.identitems)

    def __bool__(self):
        return True

    def set_special_page(self, label, target):
        "Sets a special page"
        if not label:
            raise ServerError(message="Sorry, a special page label must be given", code=9030)
        if not target:
            raise ServerError(message="Sorry, a label target must be given", code=9031)
        if isinstance(target, str) and ( '/' in target ):
                # item is a url
                item = target
        elif isinstance(target, str) and ( ',' in target ) and not (target.split(',')[1].isdigit()):
            if len(target.split(',')) == 2:
                # item points to a subproject label
                item = target
        else:
            item = skiboot.make_ident(target, self._proj_ident)
            if not item:
                raise ServerError(message="Sorry, the page target is not recognised", code=9032)
        self.special_pages[label] = item


    def delete_special_page(self, label):
        "Deletes a special page"
        if label in self.special_pages:
            del self.special_pages[label]

    def _system_page(self, label):
        """Returns the system page with the given label, if not found, returns None"""
        # label must be one of the system pages
        if label not in skiboot.sys_list():
            return
        if label not in self.special_pages:
            return
        target = self.special_pages[label]
        ident = skiboot.find_ident(target, proj_ident=self._proj_ident)
        if ident is None:
            return
        return ident.item()


    def labels(self):
        "Returns the special pages dictionary as a dictionary of labels with url's and tuples"
        labels_dict = {}
        for key, val in self.special_pages.items():
            if isinstance(val, str):
                labels_dict[key] = val
            else:
                labels_dict[key] = val.to_tuple()
        return labels_dict


    def label_value(self, key):
        "Returns the url or tuple associated with the label"
        val = self.special_pages.get(key)
        if val is None:
            return
        labels_dict = {}
        if isinstance(val, str):
            return val
        else:
            return val.to_tuple()


    def _redirect_to_url(self, url, environ, call_data, page_data, lang):
        "Return status, headers, page.data() of the redirector page, with fields set to url"
        if '/' not in url:
            raise ServerError(message="Invalid target url, must contain at least one /", code=9033)
        page = self._system_page('redirector')
        if (not page) or (page.page_type != "TemplatePage"):
            page_text = "<!DOCTYPE HTML>\n<html>\n<p>Page Redirect request. Please try: <a href=%s>%s</a></p>\n</html>" % (url, url)
            return '200 OK', [('content-type', 'text/html')], [page_text.encode('ascii', 'xmlcharrefreplace')]
        # create an ErrorMessage with the url as the message
        err = ErrorMessage(message=html.escape(url))
        # import any sections
        page.import_sections()
        page.show_error(error_messages=[err])
        # update head and body parts
        if page_data:
            page.set_values(page_data)
        page.update(environ, call_data, lang)
        status, headers = page.get_status()
        return status, headers, page.data()



    def read_form_data(self, rawformdata, caller_page):
        """Reads raw form data from the environ and returns a dictionary with keys as skiboot.WidgField objects and values as
           the form values.  Where input fields have indexed names, the skiboot.WidgField object still
           has i set to empty string, but the value is given as a dictionary with indexes as keys"""
        # rawformdata is the data obtained from environ
        # form_data is a dictionary of data returned, without the caller ident
        # and after a set of checks
        if not rawformdata:
            return {}
        if not caller_page:
            return {}
        form_data = {}
        for field in rawformdata.keys():
            # get fields and values from the rawformdata and store them in form_data
            # with keys as skiboot.WidgField objects, and values as field values
            # in the case of indexed fields, the values are dictionaries
            if field == 'ident':
                continue
            if ':' not in field:
                # All widgfields have a : in them to separate widget name from field name
                raise ValidateError(message="Form data not accepted, (invalid field %s)" % (field,))
            widgfield = skiboot.make_widgfield(field)
            # get fields and values from the rawformdata and store them in form_data
            widget = caller_page.copy_widget_from_name(widgfield.s, widgfield.w)
            if widget is None:
                raise ValidateError(message="Form data not accepted, (unexpected field %s)" % (field,))
            if isinstance(rawformdata[field], list):
                # fieldvalue is a list of items
                fieldvalue = [ item.value.strip() for item in rawformdata[field] ]
            else:
                fieldvalue = rawformdata[field].value.strip()
            if widget.is_senddict(widgfield.f):
                # field sends a dictionary, must have an index appended to the name
                # this part removes the index from the field name, and creates a form value of a dictionary with the index as keys
                fieldindex = widgfield.i
                if not fieldindex:
                    raise ValidateError(message="Form data not accepted, (invalid dictionary field %s)" % (field,))
                widgfieldnoindex = widgfield._replace(i='')
                if widgfieldnoindex in form_data:
                    form_data[widgfieldnoindex][fieldindex] = fieldvalue
                else:
                    form_data[widgfieldnoindex] = {fieldindex:fieldvalue}
            else:
                if widgfield.i:
                    raise ValidateError(message="Form data not accepted, (unexpected dictionary field %s)" % (field,))
                form_data[widgfield] = fieldvalue
        return form_data


    def _url_not_found(self, environ, path, lang):
        "Used to return the url not found page"
        page = self._system_page("url_not_found")
        page_content = "<!DOCTYPE HTML>\n<html>\nERROR:UNKNOWN URL\n</html>".encode('ascii', 'xmlcharrefreplace')
        if not page:
            return '404 Not Found', [('content-type', 'text/html')], [page_content]
        if page.page_type == "TemplatePage":
            # create an ErrorMessage with the path as the message
            err = ErrorMessage(message=html.escape(path))
            # import any sections
            page.import_sections()
            page.show_error(error_messages=[err])
        if (page.page_type == "TemplatePage") or (page.page_type == "FilePage"):
            # update head and body parts
            page.update(environ, {}, lang)
            status, headers = page.get_status()
            page_data = page.data()
            if page_data:
                return '404 Not Found', headers, page_data
        # if any other type of page, or no page content, then return this
        return '404 Not Found', [('content-type', 'text/html')], [page_content]


    def respond(self, environ):
        "called from the project top script"
        # get cookies
        try:
            cookiemorsals = cookies.SimpleCookie(environ["HTTP_COOKIE"])
        except Exception:
            cookiemorsals = None
        if cookiemorsals:
            received_cookies = {item:m.value for item,m in cookiemorsals.items()}
        else:
            received_cookies = {}
        if 'language' in received_cookies:
            language = received_cookies["language"]
        else:
            if "HTTP_ACCEPT_LANGUAGE" in environ:
                language_list = environ["HTTP_ACCEPT_LANGUAGE"].split(',')
                language = language_list[0]
            else:
                language = self.default_language
        lang = (language, self.default_language)
        try:
            # The tuple s_h_data is the tuple to return, start with it as None
            s_h_data = None

            if 'PATH_INFO' in environ:
                path = environ['PATH_INFO'].lower()
            else:
                raise ServerError(message="Invalid path", code=9034)

            # the path must start with this root project url
            if (path.find(self.url) != 0) and (path + "/" != self.url):
                # path does not start with the root, so send URL NOT FOUND
                return self._url_not_found(environ, path, lang)

            # This is the root project, check if the call is for a page in any sub project
            for proj, projurl in self._subproject_paths.items():
                if (path.find(projurl) == 0) or (path + "/" == projurl):
                    # this url is within a sub project
                    subproj = self.subprojects[proj]
                    s_h_data = subproj.proj_respond(environ, projurl, path, lang, received_cookies)
                    break
            else:
                # the call is for a page in this root project
                s_h_data = self.proj_respond(environ, self.url, path, lang, received_cookies)

            if s_h_data is None:
                # No page to return has been found, 
                return self._url_not_found(environ, path, lang)

            if s_h_data[2] is None:
                # No page data has been given
                return self._url_not_found(environ, path, lang)

            return s_h_data

        except ServerError as e:
            page = self._system_page("server_error")
            if (not page) or (page.page_type != "TemplatePage"):
                # return the default server error page
                return self.default_server_error_page(e)
            # import any sections
            page.import_sections()
            # show message passed by the exception
            page.show_error([e.errormessage])
            # if ServerError code, set it into the widget
            if e.code:
                if e.section:
                    page_data = {(e.section, e.widget, 'code'):str(e.code)}
                elif e.widget:
                    page_data = {(e.widget, 'code'):str(e.code)}
                elif page.default_error_widget.s:
                    page_data = {(page.default_error_widget.s, page.default_error_widget.w, 'code'):str(e.code)}
                elif page.default_error_widget.w:
                    page_data = {(page.default_error_widget.w, 'code'):str(e.code)}
                else:
                    page_data = None
                if page_data:
                    page.set_values(page_data)
            # update head and body parts
            page.update(environ, {}, lang, e.ident_list)
            status, headers = page.get_status()
            data = page.data()
            if not data:
                return self.default_server_error_page(e)
            # return page data
            return e.status, headers, data


    def default_server_error_page(self, e):
        "Given a ServerError exception, return a default status,headers,data"
        text_start = "<!DOCTYPE HTML>\n<html>\n<p>SERVER ERROR</p>\n<p>Error code : %s</p>\n" % (e.code,)
        if e.message:
            page_text = text_start + "<p>%s</p>\n</html>" % (html.escape(e.message),)
        else:
            page_text = text_start + "</html>"
        return '500 Internal Server Error', [('content-type', 'text/html')], [page_text.encode('ascii', 'xmlcharrefreplace')]


    def default_validate_error_page(self, e):
        "Given a ValidateError exception, return a default status,headers,data"
        if e.message:
            page_text = "<!DOCTYPE HTML>\n<html>\n<p>VALIDATION ERROR</p>\n<p>%s</p>\n</html>" % (html.escape(e.message),)
        else:
            page_text = "<!DOCTYPE HTML>\n<html>\n<p>VALIDATION ERROR</p>\n</html>"
        return '400 Bad Request', [('content-type', 'text/html')], [page_text.encode('ascii', 'xmlcharrefreplace')]


    def proj_respond(self, environ, projurl, path, lang, received_cookies):
        """projurl is the url of this project
           path is the url called
           Calls start call, and depending on the returned page, calls the project status_headers_data method"""

        # ident_data is a data string returned with the page ident, which consists of "project_pagenumber_identdata"
        # where project,pagenumber is the caller page ident

        caller_page = None
        ident_data = None

        try:
            # get the caller_page, from the ident field of submitted data
            # which could be None if no ident received
            rawformdata = cgi.FieldStorage(fp=environ['wsgi.input'], environ=environ)
            if rawformdata and ('ident' in rawformdata):
                # form data has been submitted - so get the ident field
                if not hasattr(rawformdata['ident'], 'value'):
                    raise ValidateError(message="Form data not accepted, caller page ident not recognised")
                else:
                    # rawformdata has 'ident' with attribute 'value'
                    # get the caller page ident, and the ident_data received from the 'ident' field

                   # Note: caller_page could belong to another project, so get it using ident.item() method
                   # which will query the right project

                    ident_parts = rawformdata['ident'].value.split('_', 2)
                    ident_items = len(ident_parts)
                    try:
                        if ident_items == 2:
                            caller_page = skiboot.Ident(ident_parts[0], int(ident_parts[1])).item()
                        elif ident_items == 3:
                            caller_page = skiboot.Ident(ident_parts[0], int(ident_parts[1])).item()
                            ident_data = ident_parts[2]
                    except Exception:
                        caller_page = None
                    if caller_page is None:
                       raise ValidateError(message="Form data not accepted, (received ident is not valid)")
                    if caller_page.page_type != 'TemplatePage':
                        raise ValidateError(message="Form data not accepted, (caller page ident is not a template page)")
        except ValidateError as e:
            page = self._system_page("validate_error")
            if (not page) or (page.page_type != "TemplatePage"):
                return self.default_validate_error_page(e)
            # import any sections
            page.import_sections()
            # show message passed by the exception
            page.show_error([e.errormessage])
            # update head and body parts
            page.update(environ, {}, lang, e.ident_list)
            status, headers = page.get_status()
            data = page.data()
            if not data:
                return self.default_validate_error_page(e)
            # return page data
            return e.status, headers, data

        # so caller_page could be either given, or could be None

        # get the called ident, could be None
        ident = self.page_ident_from_path(projurl, path)

        # now call the project start_call function
        try:
            if caller_page:
                caller_page_ident = caller_page.ident
            else:
                caller_page_ident = None
            # start_call could return a different page ident, or None
            pident, skicall = self.proj_start_call(environ,
                                                   path,
                                                   ident,
                                                   caller_page_ident,
                                                   received_cookies,
                                                   ident_data,
                                                   lang)
        except Exception:
            message = "Invalid exception in start_call function."
            if skiboot.get_debug():
                message += "\n"
                exc_type, exc_value, exc_traceback = sys.exc_info()
                str_list = traceback.format_exception(exc_type, exc_value, exc_traceback)
                for item in str_list:
                    message += item
            raise ServerError(message, code=9035)

        if pident is None:
            # URL NOT FOUND and start_call does not divert
            return None

        # pident is the ident of the diverted page or a label or url string

        # get the page from pident
        if isinstance(pident, str):
            # either a label, or url
            if '/' in pident:
                # get redirector page
                return self._redirect_to_url(pident, environ, skicall.call_data, skicall.page_data, skicall.lang)
            else:
                # no '/' in pident so must be a label
                pident = skiboot.find_ident_or_url(pident, self._proj_ident)
                if not pident:
                    raise ServerError(message="Returned page ident from start_call not recognised", code=9036)
                if isinstance(pident, str):
                    # must be a url, get redirector page
                    return self._redirect_to_url(pident, environ, skicall.call_data, skicall.page_data, skicall.lang)

        # so pident must be an ident
        if not isinstance(pident, skiboot.Ident):
            raise ServerError(message="Invalid ident returned from start_call", code=9037)

        # pident is the ident returned from start_call, may be in a different project
        page = pident.item()
        if page is None:
            raise ServerError(message="Invalid ident returned from start_call", code=9038)
        if page.page_type == 'Folder':
            page = page.default_page
            if not page:
                raise ServerError(message="Invalid ident returned from start_call", code=9039)

        # read any submitted data from rawformdata, and place in form_data
        try:
            form_data = {}
            if rawformdata and (caller_page is not None) and (page.page_type == "RespondPage"):
                form_data = self.read_form_data(rawformdata, caller_page)

            # so form_data only available if
                # rawformdata has been submitted
                # and caller_page is known, so widgets can be extracted
                # and the destination is a RespondPage

                # otherwise form_data is empty (though rawformdata is retained)
        

            # dependent on wether the requested page is in this project or a sub project,
            # call status_headers_data() to find the final page to return to the client

            # ident_list retains list of idents called during a call to ensure no circulating calls
            ident_list = []
            # initially no errors, e_list is a list of errors to be shown
            e_list = []

            if page.ident.proj != self._proj_ident:
                # page returned from start_call is in another project
                subproj = self.subprojects.get(page.ident.proj)
                return subproj.status_headers_data(skicall, environ, received_cookies, rawformdata, caller_page, page, ident_list, e_list, form_data)
                
            # call status_headers_data to return status, headers and data to the top script
            return self.status_headers_data(skicall, environ, received_cookies, rawformdata, caller_page, page, ident_list, e_list, form_data)

        except ValidateError as e:
            page = self._system_page("validate_error")
            if (not page) or (page.page_type != "TemplatePage"):
                return self.default_validate_error_page(e)
            # import any sections
            page.import_sections()
            # show message passed by the exception
            page.show_error([e.errormessage])
            # update head and body parts
            page.update(environ, skicall.call_data, skicall.lang, e.ident_list)
            status, headers = page.get_status()
            data = page.data()
            if not data:
                return self.default_validate_error_page(e)
            # return page data
            return e.status, headers, page.data()


    def proj_start_call(self, environ, path, ident, caller_ident, received_cookies, ident_data, lang):
        """Calls the appropriate project start_call function
           ident is the ident of the page being called, could be None if not recognised
           Returns new called_ident, dictionaries 'call_data', 'page_data' and new tuple lang"""

        if not caller_ident:
            tuple_caller_ident = ()
        else:
            tuple_caller_ident = caller_ident.to_tuple()

        if ident is None:
            called_ident = None
        else:
            called_ident = ident.to_tuple()

        if (ident_data is not None) and _AN.search(ident_data):
            ident_data = None

        try:
            # create the SkiCall object
            skicall = SkiCall(environ = environ,
                              path = path,
                              project = self._proj_ident,
                              rootproject = self.rootproject,
                              caller_ident = tuple_caller_ident,
                              received_cookies = received_cookies,
                              ident_data = ident_data,
                              lang = lang,
                              proj_data = self.proj_data)

            # the skicall object is changed in place, with call_data and page_data
            # being set by the users own start_call function
            new_called_ident = self.start_call(called_ident, skicall)

            # convert returned tuple to an Ident object
            if isinstance(new_called_ident, int):
                new_called_ident = (self._proj_ident, new_called_ident)
            if isinstance(new_called_ident, tuple):
                new_called_ident = skiboot.make_ident(new_called_ident, self._proj_ident)
            # could be a label
        except ServerError as e:
            raise e
        except Exception:
            if skiboot.get_debug():
                exc_type, exc_value, exc_traceback = sys.exc_info()
                str_list = traceback.format_exception(exc_type, exc_value, exc_traceback)
                message = ''
                for item in str_list:
                    message += item
                raise ServerError(message, code=9040)
            raise ServerError("Error in start_call", code=9041)
        return new_called_ident, skicall


    def status_headers_data(self, skicall, environ, received_cookies, rawformdata, caller_page, page, ident_list, e_list, form_data):
        """calls responders until it can return status, headers, page.data()"""

        try:
            while page.page_type == 'RespondPage':
                ident = page.ident
                if page.responder is None:
                    raise ServerError(message="Respond page %s does not have any responder set" % (page.url,), code=9042)
                try: 
                    page = page.call_responder(skicall, form_data, caller_page, ident_list, rawformdata)
                    if isinstance(page, str):
                        # must be a url
                        skicall.call_data.clear()
                        skicall.page_data.clear()
                        # get redirector page
                        return self._redirect_to_url(page, environ, skicall.call_data, skicall.page_data, skicall.lang)
                except PageError as ex:
                    # a jump to a page has occurred, with a list of errors
                    page = ex.page
                    if isinstance(page, str):
                        # must be a url
                        skicall.call_data.clear()
                        skicall.page_data.clear()
                        # get redirector page
                        return self._redirect_to_url(page, environ, skicall.call_data, skicall.page_data, skicall.lang)
                    if page.ident in ident_list:
                        raise ServerError(message="Invalid Failure page: can cause circulating call", code=9043)
                    # show the list of errors on the page
                    e_list = ex.e_list
                except GoTo as ex:
                    if ex.clear_submitted:
                        form_data.clear()
                    if ex.clear_page_data:
                        skicall.page_data.clear()
                    if ex.clear_errors:
                        ex.e_list = []
                    target = skiboot.find_ident_or_url(ex.target, ex.proj_ident)
                    # target is either an Ident, a URL or None
                    if not target:
                        raise ServerError(message="GoTo exception target not recognised", code=9044)
                    if isinstance(target, skiboot.Ident):
                        if target == ident:
                            raise ServerError(message="GoTo exception page ident %s invalid, can cause circulating call" % (target,), code=9045)
                        if target in ident_list:
                            raise ServerError(message="GoTo exception page ident %s invalid, can cause circulating call" % (target,), code=9046)
                        page = target.item()
                        if not page:
                            raise ServerError(message="GoTo exception page ident %s not recognised" % (target,), code=9047)
                        if page.page_type == 'Folder':
                            raise ServerError(message="GoTo exception page ident %s is a Folder, must be a page." % (target,), code=9048)
                    else:
                        # target is a URL
                        skicall.call_data.clear()
                        return self._redirect_to_url(target, environ, skicall.call_data, skicall.page_data, skicall.lang)
                    
                    # A divert to a fail page may lead to a GoTo exception which can therefore
                    # have an e_list
                    # show the list of errors on the page
                    e_list = ex.e_list

                # it is possible that a jump to a page in another project has been made
                if page.ident.proj != self._proj_ident:
                    subproj = skiboot.getproject(proj_ident=page.ident.proj)
                    return subproj.status_headers_data(skicall, environ, received_cookies, rawformdata, caller_page, page, ident_list, e_list, form_data)
                
        except (ServerError, ValidateError) as e:
            e.ident_list = ident_list
            raise e

        # the page to be returned to the client is now 'page'
        # and 'e_list' is a list of errors to be shown on it

        # call the user function end_call
        try:
            skicall.project = self._proj_ident
            skicall.proj_data = self.proj_data
            skicall.rootproject = self.rootproject
            try:
                session_string = self.end_call(page.ident.to_tuple(), page.page_type, skicall)
                if session_string:
                    # set cookie in target_page
                    page.session_cookie = "Set-Cookie", "%s=%s; Path=%s" % (skicall.project, session_string, skiboot.root_project().url)
            except FailPage as e:
                page.show_error([e.errormessage])
            finally:
                if skicall._lang_cookie:
                    page.language_cookie = skicall._lang_cookie
        except GoTo as e:
            raise ServerError("Invalid GoTo exception in end_call", code=9049)
        except Exception:
            message = "Invalid exception in end_call function."
            if skiboot.get_debug():
                message += "\n"
                exc_type, exc_value, exc_traceback = sys.exc_info()
                str_list = traceback.format_exception(exc_type, exc_value, exc_traceback)
                for item in str_list:
                    message += item
            raise ServerError(message, code=9050)

        # import any sections
        page.import_sections(skicall.page_data)
        if e_list:
            # show the list of errors on the page
            page.show_error(e_list)
        # now set the widget fields
        if skicall.page_data:
            page.set_values(skicall.page_data)
        page.update(environ, skicall.call_data, skicall.lang, ident_list)
        status, headers = page.get_status()
        return status, headers, page.data()


    def page_ident_from_path(self, projurl, path):
        """Tests if ident exists in the cache, return it, if not, call self.root.page_ident_from_path
           and cach the result, then return the ident. If no ident found, or if ident within a restricted folder, return None."""
        if path in self._paths:
            return self._paths[path]
        ident = None
        strip_path = path.strip("/")
        if not strip_path:
            pathlist = []
        else:
            pathlist = strip_path.split("/")
        strip_projurl = projurl.strip("/")
        # The projurl must be removed from the pathlist before the call to self.root.page_ident_from_path()
        if (not strip_projurl):
            # no projurl to remove
            ident = self.root.page_ident_from_path(self.identitems, pathlist)
        else:
            # strip_projurl may be something like "lib", remove the projurl from the pathlist
            projurl_list = strip_projurl.split("/")
            for item in projurl_list:
                if not pathlist:
                    # invalid call, the pathlist must start with the projurl
                    return
                if item == pathlist[0]:
                    pathlist.pop(0)
                else:
                    # invalid call, the pathlist must start with the projurl
                    return
            ident = self.root.page_ident_from_path(self.identitems, pathlist)
        if ident is not None:
            self._paths[path] = ident
        return ident


    @property
    def proj_ident(self):
        return self._proj_ident

    @property
    def subproject_paths(self):
        "property getter to return an ordered dictionary of sub project {ident:path,...}"
        if not self.rootproject:
            # sub project do not themselves contain further sub projects
            return collections.OrderedDict()
        return self._subproject_paths.copy()


    def list_of_subproject_idents(self):
        "Returns a list of subproject idents"
        return [i for i in self.subproject_paths]

    def add_project(self, proj, url=None):
        """Add a project to self, returns the url
           proj is the sub project application.
           This adds a reference to the project to the subproject_paths, returns the sub project path"""
        if not self.rootproject:
           raise ValidateError(message="Cannot add to a sub project")
        proj_id = proj.proj_ident

        # get a copy of the {proj_id:url} subproject_paths dictionary, and this projects url
        sub_paths = self._subproject_paths.copy()

        if (proj_id == self._proj_ident) or (proj_id in sub_paths):
            raise ValidateError(message="Project already exits")

        this_url = self.url

        if url is None:
            url = proj.url.strip("/").lower()
            if not url:
                url = proj_id.lower()
        else:
            url=url.strip("/").lower()
            if not url:
                raise ValidateError(message="Invalid URL passed to add_project, it must be a path longer than the root application path")
        url = "/" + url + "/"
        # Ensure url starts with this project url
        if not url.startswith(this_url):
            raise ValidateError(message="Invalid URL passed to add_project, it must be a path longer than the root application path")
        # add this ident and url to subproject_paths
        sub_paths[proj_id] = url
        # save new subproject_paths dictionary
        self._subproject_paths = collections.OrderedDict(sorted(sub_paths.items(), key=lambda t: len(t[1].strip("/").split("/")), reverse=True))
        # add the subproject to this project
        proj.rootproject = False
        proj.url = url
        proj._subproject_paths = collections.OrderedDict()
        proj.subprojects = {}
        self.subprojects[proj_id] = proj
        return url


    @property
    def root_ident(self):
        'provides a root_ident attribute'
        return skiboot.Ident(self._proj_ident, 0)

    def next_ident(self):
        "Returns next Ident available by incrementing the maximum existing ident number"
        return skiboot.Ident(self._proj_ident, self.max_ident_num+1)

    def __repr__(self):
        return "WSGIApplication(\'%s\', \'%s\', \'%s\', start_call, submit_data, end_call, \'%s\')" % (self.proj_ident, self.projectfiles, self.proj_data, self.url)


class SkiCall(object):

    def __init__(self, environ, path, project, rootproject, caller_ident, received_cookies, ident_data, lang, proj_data):

        self.environ = environ
        self.path = path
        self.project = project
        self.rootproject = rootproject
        self.caller_ident = caller_ident
        self.received_cookies = received_cookies
        self.ident_data = ident_data
        self._lang = lang
        self._lang_cookie = None
        self.proj_data = proj_data

        self._projectfiles = skiboot.projectfiles(project)

        self.ident_list = []
        self.submit_list = []
        self.submit_dict = {}
        self.call_data = {}
        self.page_data = {}


    @property
    def projectfiles(self):
        "Returns the projectfiles string"
        return self._projectfiles

    @property
    def lang(self):
        "Returns the lang tuple"
        return self._lang

    def get_language(self):
        "Returns the language string"
        return self._lang[0]

    def set_language(self, language):
        "Sets the language string and creates a language cookie with a persistance of 30 days"
        if language:
            self._lang = (language, self._lang[1])
            self._lang_cookie = "Set-Cookie", "language=%s; Path=%s; Max-Age=2592000" % (language_string, skiboot.root_project().url)

    language = property(get_language, set_language)

    @property
    def accesstextblocks(self):
        "Returns the project instance of the AccessTextBlocks class"
        this_project = skiboot.getproject(proj_ident=self.project)
        return this_project.textblocks

    def textblock(self, textref, project=None):
        """This method returns the textblock text, given a textblock reference string,
           If project is not given assumes this project, if given, project must exist as either the root,
           or a sub project of the root.
           If no textblock is found, returns None."""
        if project is None:
            project = self.project
        proj = skiboot.getproject(project)
        if proj is None:
            return
        return proj.textblocks.get_text(textref, self.lang)

    def label_value(self, label, project=None):
        """Given a label, returns the associated ident or URL
           If project is not given assumes this project, if given, project must exist as either the root,
           or a sub project of the root.
           If no label is found, returns None."""
        if project is None:
            project = self.project
        proj = skiboot.getproject(project)
        if proj is None:
            return
        return proj.label_value(label)


    def projectpaths(self):
        """Returns a dictionary of project name : project path

           This method returns a dictionary of project names as keys with the project paths as values."""
        all_projects = skiboot.project_register()
        return {proj_ident:proj.url for proj_ident, proj in all_projects.items()}


